@@ -9,8 +9,8 @@ import os
 import re
 
 from xpra.client.client_widget_base import ClientWidgetBase
-from xpra.os_util import bytestostr, PYTHON2, PYTHON3, OSX, WIN32
-from xpra.util import typedict, envbool, WORKSPACE_UNSET, WORKSPACE_NAMES
+from xpra.os_util import bytestostr, PYTHON2, PYTHON3, OSX, WIN32, is_Wayland
+from xpra.util import typedict, envbool, envint, WORKSPACE_UNSET, WORKSPACE_NAMES
 from xpra.log import Logger
 
 log = Logger("window")
@@ -22,13 +22,15 @@ keylog = Logger("keyboard")
 metalog = Logger("metadata")
 geomlog = Logger("geometry")
 iconlog = Logger("icon")
+alphalog = Logger("alpha")
 
 
-REPAINT_ALL = os.environ.get("XPRA_REPAINT_ALL", "")
 SIMULATE_MOUSE_DOWN = envbool("XPRA_SIMULATE_MOUSE_DOWN", True)
 PROPERTIES_DEBUG = [x.strip() for x in os.environ.get("XPRA_WINDOW_PROPERTIES_DEBUG", "").split(",")]
 AWT_DIALOG_WORKAROUND = envbool("XPRA_AWT_DIALOG_WORKAROUND", WIN32)
 SET_SIZE_CONSTRAINTS = envbool("XPRA_SET_SIZE_CONSTRAINTS", True)
+DEFAULT_GRAVITY = envint("XPRA_DEFAULT_GRAVITY", 0)
+OVERRIDE_GRAVITY = envint("XPRA_OVERRIDE_GRAVITY", 0)
 
 
 class ClientWindowBase(ClientWidgetBase):
@@ -65,6 +67,7 @@ class ClientWindowBase(ClientWidgetBase):
         self._sticky = False
         self._iconified = False
         self._focused = False
+        self.window_gravity = OVERRIDE_GRAVITY or DEFAULT_GRAVITY
         self.border = border
         self.cursor_data = None
         self.default_cursor_data = default_cursor_data
@@ -72,6 +75,7 @@ class ClientWindowBase(ClientWidgetBase):
         self.button_state = {}
         self.pixel_depth = pixel_depth      #0 for default
         self.window_offset = None           #actual vs reported coordinates
+        self.pending_refresh = []
 
         self.init_window(metadata)
         self.setup_window(bw, bh)
@@ -86,14 +90,13 @@ class ClientWindowBase(ClientWidgetBase):
         # used for only sending focus events *after* the window is mapped:
         self._been_mapped = False
         self._override_redirect_windows = []
-        if "workspace" in self._client_properties:
-            workspace = self._client_properties.get("workspace")
-            if workspace is not None:
-                workspacelog("workspace from client properties: %s", workspace)
-                #client properties override application specified workspace value on init only:
-                metadata["workspace"] = int(workspace)
+        workspace = typedict(self._client_properties).intget("workspace", None)
+        if workspace is not None:
+            workspacelog("workspace from client properties: %s", workspace)
+            #client properties override application specified workspace value on init only:
+            metadata["workspace"] = int(workspace)
         self._window_workspace = WORKSPACE_UNSET        #will get set in set_metadata if present
-        self._desktop_workspace = self.get_desktop_workspace()
+        self._desktop_workspace = self.get_desktop_workspace()  #pylint: disable=assignment-from-none
         def wn(w):
             return WORKSPACE_NAMES.get(w, w)
         workspacelog("init_window(..) workspace=%s, current workspace=%s",
@@ -101,6 +104,9 @@ class ClientWindowBase(ClientWidgetBase):
         if self.max_window_size and b"size-constraints" not in metadata:
             #this ensures that we will set size-constraints and honour max_window_size:
             metadata[b"size-constraints"] = {}
+        #initialize gravity early:
+        sc = typedict(metadata.dictget("size-constraints", {}))
+        self.window_gravity = OVERRIDE_GRAVITY or sc.intget("gravity", DEFAULT_GRAVITY)
 
 
     def get_desktop_workspace(self):
@@ -118,6 +124,7 @@ class ClientWindowBase(ClientWidgetBase):
         self._backing = self.make_new_backing(backing_class, w, h, bw, bh)
         self._backing.border = self.border
         self._backing.default_cursor_data = self.default_cursor_data
+        self._backing.gravity = self.window_gravity
         return self._backing._backing
 
 
@@ -187,12 +194,23 @@ class ClientWindowBase(ClientWidgetBase):
 
     def update_metadata(self, metadata):
         metalog("update_metadata(%s)", metadata)
+        if self._client.readonly:
+            metadata.update(self._force_size_constraint(*self._size))
         self._metadata.update(metadata)
         try:
             self.set_metadata(metadata)
         except Exception:
             metalog.warn("failed to set window metadata to '%s'", metadata, exc_info=True)
 
+    def _force_size_constraint(self, *size):
+        return {
+            b"size-constraints" : {
+                b"maximum-size" : size,
+                b"minimum-size" : size,
+                b"base-size" : size,
+                }
+            }
+
     def set_metadata(self, metadata):
         metalog("set_metadata(%s)", metadata)
         debug_props = [x for x in PROPERTIES_DEBUG if x in metadata.keys()]
@@ -211,28 +229,30 @@ class ClientWindowBase(ClientWidgetBase):
                     #perform metadata variable substitutions:
                     #full of py3k unicode headaches that don't need to be
                     default_values = {
-                                      "title"           : "<untitled window>",
-                                      "client-machine"  : "<unknown machine>",
-                                      "windowid"        : str(self._id),
-                                      }
+                        "title"           : "<untitled window>",
+                        "client-machine"  : "<unknown machine>",
+                        "windowid"        : str(self._id),
+                        "server-machine"  : getattr(self._client, "_remote_hostname", None) or "<unknown machine>",
+                        "server-display"  : getattr(self._client, "_remote_display", None) or "<unknown display>",
+                        }
                     def metadata_replace(match):
                         atvar = match.group(0)          #ie: '@title@'
                         var = atvar[1:len(atvar)-1]     #ie: 'title'
+                        if not var:
+                            #atvar = "@@"
+                            return "@"
                         default_value = default_values.get(var, "<unknown %s>" % var)
-                        value = self._metadata.strget(var, default_value)
-                        if PYTHON2:
-                            value = value.decode("utf-8")
-                        return value
+                        value = self._metadata.bytesget(var, default_value)
+                        try:
+                            return value.decode("utf-8")
+                        except UnicodeDecodeError:
+                            return str(value)
                     title = re.sub(r"@[\w\-]*@", metadata_replace, title)
-                if PYTHON2:
-                    utf8_title = title.encode("utf-8")
-                else:
-                    utf8_title = title
             except Exception as e:
                 log.error("Error parsing window title:")
                 log.error(" %s", e)
-                utf8_title = b""
-            self.set_title(utf8_title)
+                title = b""
+            self.set_title(title)
 
         if b"icon-title" in metadata:
             icon_title = metadata.strget("icon-title")
@@ -242,7 +262,7 @@ class ClientWindowBase(ClientWidgetBase):
             self.reset_icon()
 
         if b"size-constraints" in metadata:
-            sc = typedict(metadata.dictget("size-constraints"))
+            sc = typedict(metadata.dictget("size-constraints", {}))
             self.size_constraints = sc
             self._set_initial_position = sc.boolget("set-initial-position", self._set_initial_position)
             self.set_size_constraints(sc, self.max_window_size)
@@ -285,7 +305,7 @@ class ClientWindowBase(ClientWidgetBase):
         if b"has-alpha" in metadata:
             new_alpha = metadata.boolget("has-alpha")
             if new_alpha!=self._has_alpha:
-                l = metalog
+                l = alphalog
                 if PYTHON3 and not WIN32:
                     #win32 without opengl can't do transparency,
                     #so it triggers too many warnings
@@ -368,22 +388,22 @@ class ClientWindowBase(ClientWidgetBase):
             self.set_bypass_compositor(metadata.intget("bypass-compositor"))
 
         if b"strut" in metadata:
-            self.set_strut(metadata.dictget("strut"))
+            self.set_strut(metadata.dictget("strut", {}))
 
         if b"fullscreen-monitors" in metadata:
             self.set_fullscreen_monitors(metadata.intlistget("fullscreen-monitors"))
 
         if b"shape" in metadata:
-            self.set_shape(metadata.dictget("shape"))
+            self.set_shape(metadata.dictget("shape", {}))
 
         if b"command" in metadata:
             self.set_command(metadata.strget("command"))
 
-        if b"menu" in metadata:
-            self.set_menu(metadata.dictget("menu"))
+        if b"x11-property" in metadata:
+            self.set_x11_property(*metadata.listget("x11-property"))
 
 
-    def set_menu(self, menu):
+    def set_x11_property(self, *x11_property):
         pass
 
     def set_command(self, command):
@@ -414,12 +434,12 @@ class ClientWindowBase(ClientWidgetBase):
         geomlog("set_size_constraints(%s, %s)", size_constraints, max_window_size)
         hints = typedict()
         client = self._client
-        for (a, h1, h2) in [
+        for (a, h1, h2) in (
             (b"maximum-size", b"max_width", b"max_height"),
             (b"minimum-size", b"min_width", b"min_height"),
             (b"base-size", b"base_width", b"base_height"),
             (b"increment", b"width_inc", b"height_inc"),
-            ]:
+            ):
             v = size_constraints.intpair(a)
             geomlog("intpair(%s)=%s", a, v)
             if v:
@@ -442,10 +462,10 @@ class ClientWindowBase(ClientWidgetBase):
                         continue
                 hints[h1], hints[h2] = sv1, sv2
         if not OSX:
-            for (a, h) in [
+            for (a, h) in (
                 (b"minimum-aspect-ratio", b"min_aspect"),
                 (b"maximum-aspect-ratio", b"max_aspect"),
-                ]:
+                ):
                 v = size_constraints.intpair(a)
                 if v:
                     v1, v2 = v
@@ -495,7 +515,7 @@ class ClientWindowBase(ClientWidgetBase):
             #save them so the window hooks can use the last value used:
             self.geometry_hints = hints
             self.apply_geometry_hints(hints)
-        except:
+        except Exception:
             geomlog("set_size_constraints%s", (size_constraints, max_window_size), exc_info=True)
             geomlog.error("Error setting window hints:")
             for k,v in hints.items():
@@ -503,8 +523,11 @@ class ClientWindowBase(ClientWidgetBase):
             geomlog.error(" from size constraints:")
             for k,v in size_constraints.items():
                 geomlog.error(" %s=%s", k, v)
-        #TODO: handle gravity
-        #gravity = size_metadata.get("gravity")
+        self.window_gravity = OVERRIDE_GRAVITY or size_constraints.intget("gravity", DEFAULT_GRAVITY)
+        b = self._backing
+        if b:
+            b.gravity = self.window_gravity
+
 
     def set_window_type(self, window_types):
         hints = 0
@@ -606,7 +629,7 @@ class ClientWindowBase(ClientWidgetBase):
         if b:
             b.toggle()
             log("magic_key%s border=%s", args, b)
-            self.queue_draw(0, 0, *self._size)
+            self.queue_draw_area(0, 0, *self._size)
 
     def refresh_window(self, *args):
         log("refresh_window(%s) wid=%s", args, self._id)
@@ -626,27 +649,34 @@ class ClientWindowBase(ClientWidgetBase):
             from xpra.client.window_backing_base import fire_paint_callbacks
             fire_paint_callbacks(callbacks, -1, "no backing")
             return
-        def after_draw_refresh(success, message=""):
-            plog("after_draw_refresh(%s, %s) %sx%s at %sx%s encoding=%s, options=%s",
-                 success, message, width, height, x, y, coding, options)
-            if success<=0:
-                return
-            backing = self._backing
-            if backing and backing.draw_needs_refresh:
-                if REPAINT_ALL=="1" or self._client.xscale!=1 or self._client.yscale!=1:
-                    rw, rh = self.get_size()
-                    rx, ry = 0, 0
-                else:
-                    rx, ry, rw, rh = self._client.srect(x, y, width, height)
-                if self.window_offset:
-                    rx += self.window_offset[0]
-                    ry += self.window_offset[1]
-                self.idle_add(self.queue_draw, rx, ry, rw, rh)
         #only register this callback if we actually need it:
         if backing.draw_needs_refresh:
-            callbacks.append(after_draw_refresh)
+            if not backing.repaint_all:
+                self.pending_refresh.append((x, y, width, height))
+            if options.intget("flush", 0)==0:
+                callbacks.append(self.after_draw_refresh)
         backing.draw_region(x, y, width, height, coding, img_data, rowstride, options, callbacks)
 
+    def after_draw_refresh(self, success, message=""):
+        plog("after_draw_refresh(%s, %s) pending_refresh=%s",
+             success, message, self.pending_refresh)
+        backing = self._backing
+        if not backing:
+            return
+        if backing.repaint_all or self._client.xscale!=1 or self._client.yscale!=1 or is_Wayland():
+            #easy: just repaint the whole window:
+            rw, rh = self.get_size()
+            self.idle_add(self.queue_draw_area, 0, 0, rw, rh)
+            return
+        pr = self.pending_refresh
+        self.pending_refresh = []
+        for x, y, w, h in pr:
+            rx, ry, rw, rh = self._client.srect(x, y, w, h)
+            #if self.window_offset:
+            #    rx -= self.window_offset[0]
+            #    ry -= self.window_offset[1]
+            self.idle_add(self.queue_draw_area, rx, ry, rw, rh)
+
     def eos(self):
         """ Note: this runs from the draw thread (not UI thread) """
         backing = self._backing
@@ -660,7 +690,7 @@ class ClientWindowBase(ClientWidgetBase):
         #with normal windows, we just queue a draw request
         #and let the expose event paint the spinner
         w, h = self.get_size()
-        self.queue_draw(0, 0, w, h)
+        self.queue_draw_area(0, 0, w, h)
 
     def can_have_spinner(self):
         if self._backing is None:
@@ -724,7 +754,7 @@ class ClientWindowBase(ClientWidgetBase):
         #overriden in GTKClientWindowBase
         return self._id
 
-    def do_motion_notify_event(self, event):
+    def _do_motion_notify_event(self, event):
         if self._client.readonly or self._client.server_readonly or not self._client.server_pointer:
             return
         pointer, relative_pointer, modifiers, buttons = self._pointer_modifiers(event)
