@@ -15,12 +15,12 @@ from xpra.platform.gui import ready as gui_ready, get_wm_name, get_session_type,
 from xpra.version_util import full_version_str
 from xpra.net import compression, packet_encoding
 from xpra.child_reaper import reaper_cleanup
-from xpra.os_util import platform_name, bytestostr, strtobytes, BITS
+from xpra.os_util import platform_name, bytestostr, strtobytes, BITS, is_Wayland
 from xpra.util import (
     std, envbool, envint, typedict, updict, repr_ellipsized,
     XPRA_AUDIO_NOTIFICATION_ID, XPRA_DISCONNECT_NOTIFICATION_ID,
     )
-from xpra.exit_codes import EXIT_FAILURE, EXIT_OK
+from xpra.exit_codes import EXIT_CONNECTION_FAILED, EXIT_OK, EXIT_CONNECTION_LOST
 from xpra.version_util import get_version_info_full, get_platform_info
 from xpra.client import mixin_features
 from xpra.log import Logger
@@ -97,14 +97,15 @@ class UIXpraClient(ClientBaseClass):
         #same for tray:
         self.tray = None
         for c in CLIENT_BASES:
+            log("calling %s.__init__()", c)
             c.__init__(self)
         try:
             pinfo = get_platform_info()
             osinfo = "%s" % platform_name(sys.platform, pinfo.get("linux_distribution") or pinfo.get("sysrelease", ""))
             log.info(" running on %s", osinfo)
-        except:
+        except Exception:
             log("platform name error:", exc_info=True)
-        wm = get_wm_name()
+        wm = get_wm_name()      #pylint: disable=assignment-from-none
         if wm:
             log.info(" window manager is '%s'", wm)
 
@@ -121,7 +122,7 @@ class UIXpraClient(ClientBaseClass):
         self.readonly = False
         self.xsettings_enabled = False
         self.server_start_new_commands = False
-        self.xdg_menu = {}
+        self.server_xdg_menu = None
         self.start_new_commands  = []
         self.start_child_new_commands  = []
 
@@ -155,7 +156,7 @@ class UIXpraClient(ClientBaseClass):
         self._on_server_setting_changed = {}
 
 
-    def init(self, opts, extra_args=[]):
+    def init(self, opts, extra_args=()):
         """ initialize variables from configuration """
         for c in CLIENT_BASES:
             log("init: %s", c)
@@ -194,14 +195,14 @@ class UIXpraClient(ClientBaseClass):
             self.init_opengl(opts.opengl)
 
         if ClientExtras is not None:
-            self.client_extras = ClientExtras(self, opts)
+            self.client_extras = ClientExtras(self, opts)   #pylint: disable=not-callable
 
         if opts.start or opts.start_child:
             from xpra.scripts.main import strip_defaults_start_child
             from xpra.scripts.config import make_defaults_struct
             defaults = make_defaults_struct()
-            self.start_new_commands  = strip_defaults_start_child(opts.start, defaults.start)
-            self.start_child_new_commands  = strip_defaults_start_child(opts.start_child, defaults.start_child)
+            self.start_new_commands  = strip_defaults_start_child(opts.start, defaults.start)   #pylint: disable=no-member
+            self.start_child_new_commands  = strip_defaults_start_child(opts.start_child, defaults.start_child) #pylint: disable=no-member
 
         if MOUSE_DELAY_AUTO:
             try:
@@ -238,7 +239,7 @@ class UIXpraClient(ClientBaseClass):
             log("UIXpraClient.cleanup() calling %s.cleanup()", type(x))
             try:
                 x.cleanup()
-            except:
+            except Exception:
                 log.error("error on %s cleanup", type(x), exc_info=True)
         #the protocol has been closed, it is now safe to close all the windows:
         #(cleaner and needed when we run embedded in the client launcher)
@@ -306,12 +307,19 @@ class UIXpraClient(ClientBaseClass):
     # trigger notifications on disconnection,
     # and wait before actually exiting so the notification has a chance of being seen
     def server_disconnect_warning(self, reason, *info):
-        body = "\n".join(info)
-        self.may_notify(XPRA_DISCONNECT_NOTIFICATION_ID,
-                        "Xpra Session Disconnected: %s" % reason, body, icon_name="disconnected")
-        self.exit_code = EXIT_FAILURE
-        delay = NOTIFICATION_EXIT_DELAY*mixin_features.notifications
-        self.timeout_add(delay*1000, XpraClientBase.server_disconnect_warning, self, reason, *info)
+        if self.exit_code is None:
+            body = "\n".join(info)
+            if self.server_capabilities:
+                title = "Xpra Session Disconnected: %s" % reason
+                self.exit_code = EXIT_CONNECTION_LOST
+            else:
+                title = "Connection Failed: %s" % reason
+                self.exit_code = EXIT_CONNECTION_FAILED
+            self.may_notify(XPRA_DISCONNECT_NOTIFICATION_ID,
+                            title, body, icon_name="disconnected")
+            #show text notification then quit:
+            delay = NOTIFICATION_EXIT_DELAY*mixin_features.notifications
+            self.timeout_add(delay*1000, XpraClientBase.server_disconnect_warning, self, reason, *info)
         self.cleanup()
 
     def server_disconnect(self, reason, *info):
@@ -331,7 +339,7 @@ class UIXpraClient(ClientBaseClass):
         caps["session-type"] = get_session_type()
         #don't try to find the server uuid if this platform cannot run servers..
         #(doing so causes lockups on win32 and startup errors on osx)
-        if MMAP_SUPPORTED:
+        if MMAP_SUPPORTED and not is_Wayland():
             #we may be running inside another server!
             try:
                 from xpra.server.server_uuid import get_uuid
@@ -340,6 +348,7 @@ class UIXpraClient(ClientBaseClass):
                 pass
         for x in (#generic feature flags:
             "notify-startup-complete", "wants_events", "setting-change",
+            "xdg-menu-update",
             ):
             caps[x] = True
         #FIXME: the messy bits without proper namespace:
@@ -391,7 +400,7 @@ class UIXpraClient(ClientBaseClass):
         self.server_pointer = c.boolget("pointer", True)
         self.server_start_new_commands = c.boolget("start-new-commands")
         if self.server_start_new_commands:
-            self.xdg_menu = c.dictget("xdg-menu")
+            self.server_xdg_menu = c.dictget("xdg-menu", None)
         if self.start_new_commands or self.start_child_new_commands:
             if self.server_start_new_commands:
                 self.after_handshake(self.send_start_new_commands)
@@ -426,7 +435,7 @@ class UIXpraClient(ClientBaseClass):
             proxy_release = c.strget("proxy.platform.release")
             proxy_version = c.strget("proxy.version")
             proxy_version = c.strget("proxy.build.version", proxy_version)
-            proxy_distro = c.strget("linux_distribution")
+            proxy_distro = c.strget("proxy.linux_distribution")
             msg = "via: %s proxy version %s" % (
                 platform_name(proxy_platform, proxy_distro or proxy_release),
                 std(proxy_version or "unknown")
@@ -443,7 +452,7 @@ class UIXpraClient(ClientBaseClass):
         #keyboard:
         c = self.server_capabilities
         if self.keyboard_helper:
-            modifier_keycodes = c.dictget("modifier_keycodes")
+            modifier_keycodes = c.dictget("modifier_keycodes", {})
             if modifier_keycodes:
                 self.keyboard_helper.set_modifier_mappings(modifier_keycodes)
         self.key_repeat_delay, self.key_repeat_interval = c.intpair("key_repeat", (-1,-1))
@@ -465,7 +474,7 @@ class UIXpraClient(ClientBaseClass):
         for cb, args in oh:
             try:
                 cb(*args)
-            except:
+            except Exception:
                 log.error("Error processing handshake callback %s", cb, exc_info=True)
 
     def after_handshake(self, cb, *args):
@@ -499,12 +508,16 @@ class UIXpraClient(ClientBaseClass):
             "sharing", "sharing-toggle", "lock", "lock-toggle",
             "start-new-commands", "client-shutdown", "webcam",
             "bandwidth-limit", "clipboard-limits",
+            "xdg-menu",
             ):
             setattr(self, "server_%s" % setting.replace("-", "_"), value)
         else:
-            log.info("unknown server setting changed: %s=%s", setting, value)
+            log.info("unknown server setting changed: %s=%s", setting, repr_ellipsized(bytestostr(value)))
             return
-        log.info("server setting changed: %s=%s", setting, value)
+        log("_process_setting_change: %s=%s", setting, value)
+        #xdg-menu is too big to log
+        if setting not in ("xdg-menu", ):
+            log.info("server setting changed: %s=%s", setting, repr_ellipsized(str(value)))
         self.server_setting_changed(setting, value)
 
     def server_setting_changed(self, setting, value):
@@ -707,15 +720,13 @@ class UIXpraClient(ClientBaseClass):
         for c in CLIENT_BASES:
             c.init_authenticated_packet_handlers(self)
         #run from the UI thread:
-        self.set_packet_handlers(self._ui_packet_handlers, {
+        self.add_packet_handlers({
             "startup-complete":     self._process_startup_complete,
             "setting-change":       self._process_setting_change,
             "control" :             self._process_control,
             })
         #run directly from the network thread:
-        self.set_packet_handlers(self._packet_handlers, {
-            "server-event":         self._process_server_event,
-            })
+        self.add_packet_handler("server-event", self._process_server_event, False)
 
 
     def process_packet(self, proto, packet):
