@@ -1,28 +1,37 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
-# Copyright (C) 2011-2018 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2011-2019 Antoine Martin <antoine@xpra.org>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os.path
 import array
-from xpra.util import iround
-from xpra.os_util import strtobytes, bytestostr, WIN32, PYTHON2
-from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_glib, import_pixbufloader, import_pango, import_cairo, import_gobject, import_pixbuf, is_gtk3
+
+from xpra.util import iround, first_time, envint
+from xpra.os_util import (
+    strtobytes, bytestostr,
+    WIN32, OSX, PYTHON2, POSIX,
+    )
+from xpra.gtk_common.gobject_compat import (
+    import_gtk, import_gdk, import_glib, import_pixbufloader,
+    import_pango, import_cairo, import_gobject, import_pixbuf, is_gtk3,
+    )
+from xpra.log import Logger
+
+log = Logger("gtk", "util")
+traylog = Logger("gtk", "tray")
+screenlog = Logger("gtk", "screen")
+alphalog = Logger("gtk", "alpha")
+
 gtk     = import_gtk()
 gdk     = import_gdk()
 pango   = import_pango()
 cairo   = import_cairo()
 gobject = import_gobject()
+glib    = import_glib()
 PixbufLoader = import_pixbufloader()
 Pixbuf = import_pixbuf()
 
-from xpra.log import Logger
-log = Logger("gtk", "util")
-traylog = Logger("gtk", "tray")
-screenlog = Logger("gtk", "screen")
-
-
 SHOW_ALL_VISUALS = False
 
 GTK_VERSION_INFO = {}
@@ -41,7 +50,6 @@ def get_gtk_version_info():
 
     if not GTK_VERSION_INFO:
         V("gobject",    gobject,    "pygobject_version")
-        glib = import_glib()
 
         if is_gtk3():
             #this isn't the actual version, (only shows as "3.0")
@@ -57,7 +65,7 @@ def get_gtk_version_info():
             V("pixbuf",     Pixbuf,     "PIXBUF_VERSION")
             def MAJORMICROMINOR(name, module):
                 try:
-                    v = [getattr(module, x) for x in ["MAJOR_VERSION", "MICRO_VERSION", "MINOR_VERSION"]]
+                    v = tuple(getattr(module, x) for x in ("MAJOR_VERSION", "MICRO_VERSION", "MINOR_VERSION"))
                     av(name, ".".join(str(x) for x in v))
                 except:
                     pass
@@ -84,25 +92,83 @@ def get_gtk_version_info():
     return GTK_VERSION_INFO.copy()
 
 
-def popup_menu_workaround(*args):
-    #only implemented with GTK2 on win32 below
-    pass
+def pixbuf_save_to_memory(pixbuf, fmt="png"):
+    buf = []
+    def save_to_memory(data, *_args, **_kwargs):
+        buf.append(strtobytes(data))
+        return True
+    if is_gtk3():
+        pixbuf.save_to_callbackv(save_to_memory, None, fmt, [], [])
+    else:
+        pixbuf.save_to_callback(save_to_memory, fmt, {}, buf)
+    return b"".join(buf)
 
 
 if is_gtk3():
     def is_realized(widget):
         return widget.get_realized()
 
-    def make_temp_window(title):
+    def x11_foreign_new(display, xid):
+        from gi.repository import GdkX11
+        return GdkX11.X11Window.foreign_new_for_display(display, xid)
+
+    def GDKWindow(parent=None, width=1, height=1, window_type=gdk.WindowType.TOPLEVEL,
+                  event_mask=0, wclass=gdk.WindowWindowClass.INPUT_OUTPUT, title=None,
+                  x=None, y=None, override_redirect=False, visual=None, **kwargs):
+        attributes_mask = 0
         attributes = gdk.WindowAttr()
-        attributes.width = 1
-        attributes.height = 1
-        attributes.title = title
-        attributes.window_type = gdk.WindowType.TOPLEVEL
-        attributes.wclass = gdk.WindowWindowClass.INPUT_OUTPUT
-        attributes.event_mask = 0
-        attributes_mask = gdk.WindowAttributesType.TITLE | gdk.WindowAttributesType.WMCLASS
-        return gdk.Window(None, attributes, attributes_mask)
+        if x is not None:
+            attributes.x = x
+            attributes_mask |= gdk.WindowAttributesType.X
+        if y is not None:
+            attributes.y = y
+            attributes_mask |= gdk.WindowAttributesType.Y
+        #attributes.type_hint = gdk.WindowTypeHint.NORMAL
+        #attributes_mask |= gdk.WindowAttributesType.TYPE_HINT
+        attributes.width = width
+        attributes.height = height
+        attributes.window_type = window_type
+        if title:
+            attributes.title = title
+            attributes_mask |= gdk.WindowAttributesType.TITLE
+        if visual:
+            attributes.visual = visual
+            attributes_mask |= gdk.WindowAttributesType.VISUAL
+        #OR:
+        attributes.override_redirect = override_redirect
+        attributes_mask |= gdk.WindowAttributesType.NOREDIR
+        #events:
+        attributes.event_mask = event_mask
+        #wclass:
+        attributes.wclass = wclass
+        mask = gdk.WindowAttributesType(attributes_mask)
+        return gdk.Window(parent, attributes, mask)
+
+    def make_temp_window(title, window_type=gdk.WindowType.TEMP):
+        return GDKWindow(title=title, window_type=window_type)
+
+    def enable_alpha(window):
+        screen = window.get_screen()
+        visual = screen.get_rgba_visual()
+        alphalog("enable_alpha(%s) screen=%s, visual=%s", window, screen, visual)
+        #we can't do alpha on win32 with plain GTK,
+        #(though we handle it in the opengl backend)
+        if WIN32:
+            l = alphalog
+        else:
+            l = alphalog.error
+        if visual is None or (not WIN32 and not screen.is_composited()):
+            l("Error: cannot handle window transparency")
+            if visual is None:
+                l(" no RGBA visual")
+            else:
+                assert not screen.is_composited()
+                l(" screen is not composited")
+            return False
+        alphalog("enable_alpha(%s) using rgba visual %s", window, visual)
+        window.set_visual(visual)
+        return True
+
 
     def get_pixbuf_from_data(rgb_data, has_alpha, w, h, rowstride):
         data = array.array('B', strtobytes(rgb_data))
@@ -134,34 +200,37 @@ if is_gtk3():
     def get_xwindow(w):
         return w.get_xid()
     def get_default_root_window():
-        return gdk.Screen.get_default().get_root_window()
+        screen = gdk.Screen.get_default()
+        if screen is None:
+            return None
+        return screen.get_root_window()
     def get_root_size():
-        if WIN32:
+        if WIN32 or (POSIX and not OSX):
             #FIXME: hopefully, we can remove this code once GTK3 on win32 is fixed?
             #we do it the hard way because the root window geometry is invalid on win32:
             #and even just querying it causes this warning:
             #"GetClientRect failed: Invalid window handle."
-            display = gdk.Display.get_default()
-            n = display.get_n_screens()
-            w, h = 0, 0
-            for i in range(n):
-                screen = display.get_screen(i)
-                w += screen.get_width()
-                h += screen.get_height()
+            screen = gdk.Screen.get_default()
+            if screen is None:
+                return 1920, 1024
+            w = screen.get_width()
+            h = screen.get_height()
         else:
             #the easy way for platforms that work out of the box:
             root = get_default_root_window()
             w, h = root.get_geometry()[2:4]
         if w<=0 or h<=0 or w>32768 or h>32768:
-            log.warn("Warning: Gdk returned invalid root window dimensions: %ix%i", w, h)
-            w, h = 1920, 1080
-            log.warn(" using %ix%i instead", w, h)
+            if first_time("Gtk root window dimensions"):
+                log.warn("Warning: Gdk returned invalid root window dimensions: %ix%i", w, h)
+                w, h = 1920, 1080
+                log.warn(" using %ix%i instead", w, h)
         return w, h
 
     keymap_get_for_display  = gdk.Keymap.get_for_display
 
     def get_default_cursor():
-        return gdk.Cursor.new(gdk.CursorType.X_CURSOR)
+        display = gdk.Display.get_default()
+        return gdk.Cursor.new_from_name(display, "default")
     new_Cursor_for_display  = gdk.Cursor.new_for_display
     new_Cursor_from_pixbuf  = gdk.Cursor.new_from_pixbuf
     from gi.repository import GdkPixbuf     #@UnresolvedImport
@@ -176,7 +245,6 @@ if is_gtk3():
     def pixbuf_new_from_data(*args):
         args = list(args)+[None, None]
         return GdkPixbuf.Pixbuf.new_from_data(*args)
-    get_default_keymap      = gdk.Keymap.get_default
     display_get_default     = gdk.Display.get_default
     screen_get_default      = gdk.Screen.get_default
     cairo_set_source_pixbuf = gdk.cairo_set_source_pixbuf
@@ -191,6 +259,11 @@ if is_gtk3():
     WIN_POS_CENTER  = gtk.WindowPosition.CENTER
     RESPONSE_CANCEL = gtk.ResponseType.CANCEL
     RESPONSE_OK     = gtk.ResponseType.OK
+    RESPONSE_REJECT = gtk.ResponseType.REJECT
+    RESPONSE_ACCEPT = gtk.ResponseType.ACCEPT
+    RESPONSE_CLOSE  = gtk.ResponseType.CLOSE
+    RESPONSE_DELETE_EVENT = gtk.ResponseType.DELETE_EVENT
+
 
     WINDOW_STATE_WITHDRAWN  = gdk.WindowState.WITHDRAWN
     WINDOW_STATE_ICONIFIED  = gdk.WindowState.ICONIFIED
@@ -202,10 +275,14 @@ if is_gtk3():
     FILE_CHOOSER_ACTION_SAVE    = gtk.FileChooserAction.SAVE
     FILE_CHOOSER_ACTION_OPEN    = gtk.FileChooserAction.OPEN
     PROPERTY_CHANGE_MASK = gdk.EventMask.PROPERTY_CHANGE_MASK
+    FOCUS_CHANGE_MASK    = gdk.EventMask.FOCUS_CHANGE_MASK
     BUTTON_PRESS_MASK    = gdk.EventMask.BUTTON_PRESS_MASK
     BUTTON_RELEASE_MASK  = gdk.EventMask.BUTTON_RELEASE_MASK
     ENTER_NOTIFY_MASK    = gdk.EventMask.ENTER_NOTIFY_MASK
     LEAVE_NOTIFY_MASK    = gdk.EventMask.LEAVE_NOTIFY_MASK
+    SUBSTRUCTURE_MASK    = gdk.EventMask.SUBSTRUCTURE_MASK
+    STRUCTURE_MASK       = gdk.EventMask.STRUCTURE_MASK
+    EXPOSURE_MASK        = gdk.EventMask.EXPOSURE_MASK
     ACCEL_LOCKED = gtk.AccelFlags.LOCKED
     ACCEL_VISIBLE = gtk.AccelFlags.VISIBLE
     JUSTIFY_LEFT    = gtk.Justification.LEFT
@@ -223,6 +300,14 @@ if is_gtk3():
     WINDOW_POPUP    = gtk.WindowType.POPUP
     WINDOW_TOPLEVEL = gtk.WindowType.TOPLEVEL
 
+    GDKWINDOW_TEMP     = gdk.WindowType.TEMP
+    GDKWINDOW_TOPLEVEL = gdk.WindowType.TOPLEVEL
+    GDKWINDOW_CHILD    = gdk.WindowType.CHILD
+    GDKWINDOW_FOREIGN  = gdk.WindowType.FOREIGN
+
+    CLASS_INPUT_OUTPUT = gdk.WindowWindowClass.INPUT_OUTPUT
+    CLASS_INPUT_ONLY = gdk.WindowWindowClass.INPUT_ONLY
+
     LSB_FIRST       = gdk.ByteOrder.LSB_FIRST
     MSB_FIRST       = gdk.ByteOrder.MSB_FIRST
     STATIC_GRAY     = gdk.VisualType.STATIC_GRAY
@@ -312,12 +397,10 @@ if is_gtk3():
         widget.drag_get_data(context, atom, time)
 
     from gi.repository.Gtk import Clipboard     #@UnresolvedImport
-    CLIPBOARD_SELECTION = {}
-    #gtk2: uses strings:
-    for x in ("PRIMARY", "SECONDARY", "CLIPBOARD"):
-        CLIPBOARD_SELECTION[x] = getattr(gdk, "SELECTION_%s" % bytestostr(x))
     def GetClipboard(selection):
-        return Clipboard.get(CLIPBOARD_SELECTION[selection])
+        sstr = bytestostr(selection)
+        atom = getattr(gdk, "SELECTION_%s" % sstr, None) or gdk.Atom.intern(sstr, False)
+        return Clipboard.get(atom)
     def clipboard_request_contents(clipboard, target, unpack):
         target_atom = gdk.Atom.intern(bytestostr(target), False)
         clipboard.request_contents(target_atom, unpack)
@@ -364,13 +447,11 @@ if is_gtk3():
             return self.get_popup()
 
     gdk_window_process_all_updates = gdk.Window.process_all_updates
-    def gtk_main():
-        gdk.threads_init()
-        try:
-            gdk.threads_enter()
-            gtk.main()
-        finally:
-            gdk.threads_leave()
+    gtk_main = gtk.main
+
+    def popup_menu_workaround(*args):
+        #only implemented with GTK2 on win32 below
+        pass
 
     def gio_File(path):
         from gi.repository import Gio   #@UnresolvedImport
@@ -387,12 +468,41 @@ if is_gtk3():
         _, data, etag = gfile.load_contents_finish(res)
         return data, len(data), etag
 
+    def set_clipboard_data(clipboard, thevalue, _vtype="STRING"):
+        #only strings with GTK3?
+        thestring = str(thevalue)
+        clipboard.set_text(thestring, len(thestring))
+
+    def wait_for_contents(clipboard, target):
+        atom = gdk.Atom.intern(target, False)
+        return clipboard.wait_for_contents(atom)
+
+    PARAM_READABLE = gobject.ParamFlags.READABLE
+    PARAM_READWRITE = gobject.ParamFlags.READWRITE
 else:
     def get_pixbuf_from_data(rgb_data, has_alpha, w, h, rowstride):
         return gdk.pixbuf_new_from_data(rgb_data, gdk.COLORSPACE_RGB, has_alpha, 8, w, h, rowstride)
 
-    def make_temp_window(title):
-        return gdk.Window(None, 1, 1, gdk.WINDOW_TOPLEVEL, 0, gdk.INPUT_OUTPUT, title)
+    def x11_foreign_new(display, xid):
+        return gdk.window_foreign_new_for_display(display, xid)
+
+    def GDKWindow(parent=None, width=1, height=1, window_type=gdk.WINDOW_TOPLEVEL,
+                  event_mask=0, wclass=gdk.INPUT_OUTPUT, title=None, x=-1, y=-1, **kwargs):
+        return gdk.Window(parent, width, height, window_type, event_mask, wclass, title, x, y, **kwargs)
+
+    def make_temp_window(title, window_type=gdk.WINDOW_TEMP):
+        return GDKWindow(title=title, window_type=window_type)
+
+    def enable_alpha(window):
+        screen = window.get_screen()
+        rgba = screen.get_rgba_colormap()
+        alphalog("enable_alpha(%s) rgba colormap=%s", window, rgba)
+        if rgba is None:
+            log.error("Error: cannot handle window transparency, no RGBA colormap")
+            return False
+        alphalog("enable_alpha(%s) using rgba colormap %s", window, rgba)
+        window.set_colormap(rgba)
+        return True
 
     def get_pixbuf_from_window(window, x, y, w, h):
         pixbuf = gdk.Pixbuf(gdk.COLORSPACE_RGB, False, 8, w, h)
@@ -430,7 +540,6 @@ else:
     image_new_from_pixbuf   = gtk.image_new_from_pixbuf
     pixbuf_new_from_file    = gdk.pixbuf_new_from_file
     pixbuf_new_from_data    = gdk.pixbuf_new_from_data
-    get_default_keymap      = gdk.keymap_get_default
     display_get_default     = gdk.display_get_default
     screen_get_default      = gdk.screen_get_default
     window_set_default_icon = gtk.window_set_default_icon
@@ -452,6 +561,10 @@ else:
     WIN_POS_CENTER  = gtk.WIN_POS_CENTER
     RESPONSE_CANCEL = gtk.RESPONSE_CANCEL
     RESPONSE_OK     = gtk.RESPONSE_OK
+    RESPONSE_REJECT = gtk.RESPONSE_REJECT
+    RESPONSE_ACCEPT = gtk.RESPONSE_ACCEPT
+    RESPONSE_CLOSE  = gtk.RESPONSE_CLOSE
+    RESPONSE_DELETE_EVENT = gtk.RESPONSE_DELETE_EVENT
 
     WINDOW_STATE_WITHDRAWN  = gdk.WINDOW_STATE_WITHDRAWN
     WINDOW_STATE_ICONIFIED  = gdk.WINDOW_STATE_ICONIFIED
@@ -463,10 +576,14 @@ else:
     FILE_CHOOSER_ACTION_SAVE    = gtk.FILE_CHOOSER_ACTION_SAVE
     FILE_CHOOSER_ACTION_OPEN    = gtk.FILE_CHOOSER_ACTION_OPEN
     PROPERTY_CHANGE_MASK = gdk.PROPERTY_CHANGE_MASK
+    FOCUS_CHANGE_MASK    = gdk.FOCUS_CHANGE_MASK
     BUTTON_PRESS_MASK    = gdk.BUTTON_PRESS_MASK
     BUTTON_RELEASE_MASK  = gdk.BUTTON_RELEASE_MASK
     ENTER_NOTIFY_MASK    = gdk.ENTER_NOTIFY_MASK
     LEAVE_NOTIFY_MASK    = gdk.LEAVE_NOTIFY_MASK
+    SUBSTRUCTURE_MASK    = gdk.SUBSTRUCTURE_MASK
+    STRUCTURE_MASK       = gdk.STRUCTURE_MASK
+    EXPOSURE_MASK        = gdk.EXPOSURE_MASK
     ACCEL_LOCKED = gtk.ACCEL_LOCKED
     ACCEL_VISIBLE = gtk.ACCEL_VISIBLE
     JUSTIFY_LEFT    = gtk.JUSTIFY_LEFT
@@ -484,6 +601,14 @@ else:
     WINDOW_POPUP    = gtk.WINDOW_POPUP
     WINDOW_TOPLEVEL = gtk.WINDOW_TOPLEVEL
 
+    GDKWINDOW_TEMP     = gdk.WINDOW_TEMP
+    GDKWINDOW_TOPLEVEL = gdk.WINDOW_TOPLEVEL
+    GDKWINDOW_CHILD    = gdk.WINDOW_CHILD
+    GDKWINDOW_FOREIGN  = gdk.WINDOW_FOREIGN
+
+    CLASS_INPUT_OUTPUT = gdk.INPUT_OUTPUT
+    CLASS_INPUT_ONLY = gdk.INPUT_ONLY
+
     LSB_FIRST       = gdk.LSB_FIRST
     MSB_FIRST       = gdk.MSB_FIRST
     STATIC_GRAY     = gdk.VISUAL_STATIC_GRAY
@@ -627,19 +752,20 @@ else:
             mouse_in_tray_menu = False
             def check_menu_left(expected_counter):
                 if mouse_in_tray_menu:
-                    return    False
+                    return False
                 if expected_counter!=mouse_in_tray_menu_counter:
-                    return    False            #counter has changed
+                    return False            #counter has changed
                 close_cb()
+                return False
             gobject.timeout_add(OUTSIDE_TRAY_TIMEOUT, check_menu_left, mouse_in_tray_menu_counter)
         mouse_in_tray_menu_counter = 0
         mouse_in_tray_menu = False
-        traylog("popup_menu_workaround: adding events callbacks")
+        #traylog("popup_menu_workaround: adding events callbacks")
         menu.connect("enter-notify-event", enter_menu)
         menu.connect("leave-notify-event", leave_menu)
 
     def gio_File(path):
-        import gio
+        import gio      #@UnresolvedImport
         return gio.File(path=path)
 
     def query_info_async(gfile, attributes, callback, flags=0, cancellable=None):
@@ -651,12 +777,39 @@ else:
     def load_contents_finish(gfile, res):
         return gfile.load_contents_finish(res)
 
+    def set_clipboard_data(clipboard, thevalue, vtype="STRING"):
+        TEXT_TARGETS = ("UTF8_STRING", "TEXT", "STRING", "text/plain")
+        if WIN32:
+            if vtype in TEXT_TARGETS:
+                clipboard.set_text(thevalue, len(thevalue))
+            return
+        if OSX and vtype in TEXT_TARGETS:
+            clipboard.set_text(thevalue, len(thevalue))
+            return
+        value = [thevalue]
+        def get_func(clipboard, selection, info, targets):
+            log("get_func%s value=%s", (clipboard, selection, info, targets), value[0])
+            s = value[0]
+            if s:
+                selection.set(vtype, 8, s)
+            else:
+                clipboard.clear()
+        def clear_func(*args):
+            log("clear_func%s value=%s", args, value[0])
+            value[0] = ""
+        clipboard.set_with_data([(vtype, 0, 0)], get_func, clear_func)
+
+    def wait_for_contents(clipboard, target):
+        return clipboard.wait_for_contents(target)
+
+    PARAM_READABLE = gobject.PARAM_READABLE
+    PARAM_READWRITE = gobject.PARAM_READWRITE
+
 #no idea why, but trying to use the threads_init / threads_enter
 #causes deadlocks on win32:
 if WIN32:
-    def gtk_main():
-        gtk.main()
-
+    gtk_main = gtk.main
+gtk_main_quit = gtk.main_quit
 
 GRAB_STATUS_STRING = {
                       GRAB_SUCCESS          : "SUCCESS",
@@ -700,14 +853,14 @@ class TrayCheckMenuItem(gtk.CheckMenuItem):
         traylog("TrayCheckMenuItem.on_button_release_event(%s) label=%s", args, self.label)
         self.active_state = self.get_active()
         def recheck():
-            traylog("TrayCheckMenuItem: recheck() active_state=%s, get_active()=%s", self.active_state, self.get_active())
+            traylog("TrayCheckMenuItem: recheck() active_state=%s, get_active()=%s",
+                    self.active_state, self.get_active())
             state = self.active_state
             self.active_state = None
             if state is not None and state==self.get_active():
                 #toggle did not fire after the button release, so force it:
                 self.set_active(not state)
-        gobject.idle_add(recheck)
-
+        glib.idle_add(recheck)
 
 class TrayImageMenuItem(gtk.ImageMenuItem):
     """ We add a button handler to catch clicks that somehow do not
@@ -735,24 +888,24 @@ class TrayImageMenuItem(gtk.ImageMenuItem):
                 self.activate()
                 #not essential since we'll clear it before calling recheck again:
                 self._activated = False
-        gobject.idle_add(recheck)
+        glib.idle_add(recheck)
 
 
 ImageMenuItemClass = gtk.ImageMenuItem
 CheckMenuItemClass = gtk.CheckMenuItem
-def CheckMenuItem(*args, **kwargs):
+def CheckMenuItem(label):
     global CheckMenuItemClass
-    return CheckMenuItemClass(*args, **kwargs)
+    return CheckMenuItemClass(label=label)
 
 def set_use_tray_workaround(enabled):
     global CheckMenuItemClass, ImageMenuItemClass
-    if enabled:
+    if enabled and PYTHON2:
         CheckMenuItemClass = TrayCheckMenuItem
         ImageMenuItemClass = TrayImageMenuItem
     else:
         CheckMenuItemClass = gtk.CheckMenuItem
         ImageMenuItemClass = gtk.ImageMenuItem
-set_use_tray_workaround(PYTHON2)
+set_use_tray_workaround(True)
 
 
 
@@ -773,56 +926,127 @@ def get_screen_sizes(xscale=1, yscale=1):
     def swork(*workarea):
         return xs(workarea[0]), ys(workarea[1]), xs(workarea[2]), ys(workarea[3])
     display = display_get_default()
-    i=0
-    screen_sizes = []
+    if not display:
+        return ()
+    MIN_DPI = envint("XPRA_MIN_DPI", 10)
+    MAX_DPI = envint("XPRA_MIN_DPI", 500)
+    def dpi(size_pixels, size_mm):
+        if size_mm==0:
+            return 0
+        return int(size_pixels * 254 / size_mm / 10)
     n_screens = display.get_n_screens()
-    screenlog("get_screen_sizes(%f, %f) found %s screens", xscale, yscale, n_screens)
-    while i<n_screens:
-        screen = display.get_screen(i)
-        j = 0
+    get_n_monitors = getattr(display, "get_n_monitors", None)
+    if get_n_monitors:
+        #GTK 3.22: always just one screen
+        n_monitors = get_n_monitors()
+        workareas = get_workareas()
+        if workareas and len(workareas)!=n_monitors:
+            screenlog(" workareas: %s", workareas)
+            screenlog(" number of monitors does not match number of workareas!")
+            workareas = []
         monitors = []
-        workareas = []
-        #native "get_workareas()" is only valid for a single screen (but describes all the monitors)
-        #and it is only implemented on win32 right now
-        #other platforms only implement "get_workarea()" instead, which is reported against the screen
-        n_monitors = screen.get_n_monitors()
-        screenlog(" screen %s has %s monitors", i, n_monitors)
-        if n_screens==1:
-            workareas = get_workareas()
-            if workareas and len(workareas)!=n_monitors:
-                screenlog(" workareas: %s", workareas)
-                screenlog(" number of monitors does not match number of workareas!")
-                workareas = []
-        while j<screen.get_n_monitors():
-            geom = screen.get_monitor_geometry(j)
-            plug_name = ""
-            if hasattr(screen, "get_monitor_plug_name"):
-                plug_name = screen.get_monitor_plug_name(j) or ""
-            wmm = -1
-            if hasattr(screen, "get_monitor_width_mm"):
-                wmm = screen.get_monitor_width_mm(j)
-            hmm = -1
-            if hasattr(screen, "get_monitor_height_mm"):
-                hmm = screen.get_monitor_height_mm(j)
+        for j in range(n_monitors):
+            monitor = display.get_monitor(j)
+            geom = monitor.get_geometry()
+            manufacturer, model = monitor.get_manufacturer(), monitor.get_model()
+            if manufacturer and model:
+                plug_name = "%s %s" % (manufacturer, model)
+            elif manufacturer:
+                plug_name = manufacturer
+            elif model:
+                plug_name = model
+            else:
+                plug_name = "%i" % j
+            wmm, hmm = monitor.get_width_mm(), monitor.get_height_mm()
             monitor = [plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
             screenlog(" monitor %s: %s", j, monitor)
             if workareas:
                 w = workareas[j]
                 monitor += list(swork(*w))
             monitors.append(tuple(monitor))
-            j += 1
-        work_x, work_y, work_width, work_height = swork(0, 0, screen.get_width(), screen.get_height())
-        workarea = get_workarea()
+        screen = display.get_default_screen()
+        sw, sh = screen.get_width(), screen.get_height()
+        work_x, work_y, work_width, work_height = swork(0, 0, sw, sh)
+        workarea = get_workarea()   #pylint: disable=assignment-from-none
         if workarea:
-            work_x, work_y, work_width, work_height = swork(*workarea)
+            work_x, work_y, work_width, work_height = swork(*workarea)  #pylint: disable=not-an-iterable
         screenlog(" workarea=%s", workarea)
-        item = (screen.make_display_name(), xs(screen.get_width()), ys(screen.get_height()),
-                    screen.get_width_mm(), screen.get_height_mm(),
+        wmm = screen.get_width_mm()
+        hmm = screen.get_height_mm()
+        xdpi = dpi(sw, wmm)
+        ydpi = dpi(sh, hmm)
+        if xdpi<MIN_DPI or xdpi>MAX_DPI or ydpi<MIN_DPI or ydpi>MAX_DPI:
+            warn = first_time("invalid-screen-size-%ix%i" % (wmm, hmm))
+            if warn:
+                log.warn("Warning: invalid screen size %ix%imm", wmm, hmm)
+            if n_monitors>0:
+                wmm = sum(display.get_monitor(i).get_width_mm() for i in range(n_monitors))
+                hmm = sum(display.get_monitor(i).get_height_mm() for i in range(n_monitors))
+                xdpi = dpi(sw, wmm)
+                ydpi = dpi(sh, hmm)
+            if xdpi<MIN_DPI or xdpi>MAX_DPI or ydpi<MIN_DPI or ydpi>MAX_DPI:
+                #still invalid, generate one from DPI=96
+                wmm = iround(sw*254/10/96.0)
+                hmm = iround(sh*254/10/96.0)
+            if warn:
+                log.warn(" using %ix%i mm", wmm, hmm)
+        item = (screen.make_display_name(), xs(sw), ys(sh),
+                    wmm, hmm,
                     monitors,
                     work_x, work_y, work_width, work_height)
-        screenlog(" screen %s: %s", i, item)
-        screen_sizes.append(item)
-        i += 1
+        screenlog(" screen: %s", item)
+        screen_sizes = [item]
+    else:
+        i=0
+        screen_sizes = []
+        #GTK2 or GTK3<3.22:
+        screenlog("get_screen_sizes(%f, %f) found %s screens", xscale, yscale, n_screens)
+        while i<n_screens:
+            screen = display.get_screen(i)
+            j = 0
+            monitors = []
+            workareas = []
+            #native "get_workareas()" is only valid for a single screen (but describes all the monitors)
+            #and it is only implemented on win32 right now
+            #other platforms only implement "get_workarea()" instead, which is reported against the screen
+            n_monitors = screen.get_n_monitors()
+            screenlog(" screen %s has %s monitors", i, n_monitors)
+            if n_screens==1:
+                workareas = get_workareas()
+                if workareas and len(workareas)!=n_monitors:
+                    screenlog(" workareas: %s", workareas)
+                    screenlog(" number of monitors does not match number of workareas!")
+                    workareas = []
+            while j<screen.get_n_monitors():
+                geom = screen.get_monitor_geometry(j)
+                plug_name = ""
+                if hasattr(screen, "get_monitor_plug_name"):
+                    plug_name = screen.get_monitor_plug_name(j) or ""
+                wmm = -1
+                if hasattr(screen, "get_monitor_width_mm"):
+                    wmm = screen.get_monitor_width_mm(j)
+                hmm = -1
+                if hasattr(screen, "get_monitor_height_mm"):
+                    hmm = screen.get_monitor_height_mm(j)
+                monitor = [plug_name, xs(geom.x), ys(geom.y), xs(geom.width), ys(geom.height), wmm, hmm]
+                screenlog(" monitor %s: %s", j, monitor)
+                if workareas:
+                    w = workareas[j]
+                    monitor += list(swork(*w))
+                monitors.append(tuple(monitor))
+                j += 1
+            work_x, work_y, work_width, work_height = swork(0, 0, screen.get_width(), screen.get_height())
+            workarea = get_workarea()   #pylint: disable=assignment-from-none
+            if workarea:
+                work_x, work_y, work_width, work_height = swork(*workarea)  #pylint: disable=not-an-iterable
+            screenlog(" workarea=%s", workarea)
+            item = (screen.make_display_name(), xs(screen.get_width()), ys(screen.get_height()),
+                        screen.get_width_mm(), screen.get_height_mm(),
+                        monitors,
+                        work_x, work_y, work_width, work_height)
+            screenlog(" screen %s: %s", i, item)
+            screen_sizes.append(item)
+            i += 1
     return screen_sizes
 
 def get_screen_info(display, screen):
@@ -851,11 +1075,32 @@ def get_screen_info(display, screen):
         if fo:
             fontoptions = info.setdefault("fontoptions", {})
             for x,vdict in {
-                            "antialias"     : {cairo.ANTIALIAS_DEFAULT      : "default", cairo.ANTIALIAS_NONE       : "none",   cairo.ANTIALIAS_GRAY        : "gray",   cairo.ANTIALIAS_SUBPIXEL    : "subpixel"},
-                            "hint_metrics"  : {cairo.HINT_METRICS_DEFAULT   : "default", cairo.HINT_METRICS_OFF     : "off",    cairo.HINT_METRICS_ON       : "on"},
-                            "hint_style"    : {cairo.HINT_STYLE_DEFAULT     : "default", cairo.HINT_STYLE_NONE      : "none",   cairo.HINT_STYLE_SLIGHT     : "slight", cairo.HINT_STYLE_MEDIUM     : "medium", cairo.HINT_STYLE_FULL       : "full"},
-                            "subpixel_order": {cairo.SUBPIXEL_ORDER_DEFAULT : "default", cairo.SUBPIXEL_ORDER_RGB   : "RGB",    cairo.SUBPIXEL_ORDER_BGR    : "BGR",    cairo.SUBPIXEL_ORDER_VRGB   : "VRGB",   cairo.SUBPIXEL_ORDER_VBGR   : "VBGR"},
-                            }.items():
+                "antialias" : {
+                    cairo.ANTIALIAS_DEFAULT     : "default",
+                    cairo.ANTIALIAS_NONE        : "none",
+                    cairo.ANTIALIAS_GRAY        : "gray",
+                    cairo.ANTIALIAS_SUBPIXEL    : "subpixel",
+                    },
+                "hint_metrics" : {
+                    cairo.HINT_METRICS_DEFAULT  : "default",
+                    cairo.HINT_METRICS_OFF      : "off",
+                    cairo.HINT_METRICS_ON       : "on",
+                    },
+                "hint_style" : {
+                    cairo.HINT_STYLE_DEFAULT    : "default",
+                    cairo.HINT_STYLE_NONE       : "none",
+                    cairo.HINT_STYLE_SLIGHT     : "slight",
+                    cairo.HINT_STYLE_MEDIUM     : "medium",
+                    cairo.HINT_STYLE_FULL       : "full",
+                    },
+                "subpixel_order": {
+                    cairo.SUBPIXEL_ORDER_DEFAULT    : "default",
+                    cairo.SUBPIXEL_ORDER_RGB        : "RGB",
+                    cairo.SUBPIXEL_ORDER_BGR        : "BGR",
+                    cairo.SUBPIXEL_ORDER_VRGB       : "VRGB",
+                    cairo.SUBPIXEL_ORDER_VBGR       : "VBGR",
+                    },
+                }.items():
                 fn = getattr(fo, "get_"+x)
                 val = fn()
                 fontoptions[x] = vdict.get(val, val)
@@ -865,24 +1110,37 @@ def get_screen_info(display, screen):
     def visual(name, v):
         if not v:
             return
-        for x, vdict in {"bits_per_rgb" : {},
-                         "byte_order"   : {LSB_FIRST    : "LSB", MSB_FIRST  : "MSB"},
-                         "colormap_size": {},
-                         "depth"        : {},
-                         "red_pixel_details"    : {},
-                         "green_pixel_details"  : {},
-                         "blue_pixel_details"   : {},
-                         "visual_type"  : {STATIC_GRAY : "STATIC_GRAY", GRAYSCALE : "GRAYSCALE",  STATIC_COLOR : "STATIC_COLOR", PSEUDO_COLOR : "PSEUDO_COLOR", TRUE_COLOR : "TRUE_COLOR", DIRECT_COLOR : "DIRECT_COLOR"},
-                         }.items():
+        for x, vdict in {
+            "bits_per_rgb"          : {},
+            "byte_order"            : {
+                LSB_FIRST : "LSB",
+                MSB_FIRST : "MSB",
+            },
+            "colormap_size"         : {},
+            "depth"                 : {},
+            "red_pixel_details"     : {},
+            "green_pixel_details"   : {},
+            "blue_pixel_details"    : {},
+            "visual_type"           : {
+                STATIC_GRAY    : "STATIC_GRAY",
+                GRAYSCALE      : "GRAYSCALE",
+                STATIC_COLOR   : "STATIC_COLOR",
+                PSEUDO_COLOR   : "PSEUDO_COLOR",
+                TRUE_COLOR     : "TRUE_COLOR",
+                DIRECT_COLOR   : "DIRECT_COLOR",
+                },
+            }.items():
             val = None
             try:
-                val = getattr(v, x.replace("visual_"))  #ugly workaround for "visual_type" -> "type" for GTK2...
-            except:
+                #ugly workaround for "visual_type" -> "type" for GTK2...
+                val = getattr(v, x.replace("visual_", ""))
+            except AttributeError:
                 try:
                     fn = getattr(v, "get_"+x)
-                    val = fn()
-                except:
+                except AttributeError:
                     pass
+                else:
+                    val = fn()
             if val is not None:
                 vinfo.setdefault(name, {})[x] = vdict.get(val, val)
     try:
@@ -1067,7 +1325,12 @@ def get_icon_from_file(filename):
     return pixbuf
 
 
-def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32, default=False, min_size=None, label_color=None):
+def set_tooltip_text(widget, text):
+    widget.set_tooltip_text(text)
+
+
+def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32,
+                default=False, min_size=None, label_color=None, label_font=None):
     button = gtk.Button(title)
     settings = button.get_settings()
     settings.set_property('gtk-button-images', True)
@@ -1076,7 +1339,7 @@ def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32,
             icon = scaled_image(icon, icon_size)
         button.set_image(icon)
     if tooltip:
-        button.set_tooltip_text(tooltip)
+        set_tooltip_text(button, tooltip)
     if min_size:
         button.set_size_request(min_size, min_size)
     if clicked_callback:
@@ -1086,11 +1349,18 @@ def imagebutton(title, icon, tooltip=None, clicked_callback=None, icon_size=32,
             button.set_can_default(True)
         else:
             button.set_flags(gtk.CAN_DEFAULT)
-    if label_color:
-        alignment = button.get_children()[0]
-        b_hbox = alignment.get_children()[0]
-        label = b_hbox.get_children()[1]
-        label.modify_fg(STATE_NORMAL, label_color)
+    if label_color or label_font:
+        try:
+            alignment = button.get_children()[0]
+            b_hbox = alignment.get_children()[0]
+            label = b_hbox.get_children()[1]
+        except IndexError:
+            pass
+        else:
+            if label_color:
+                label.modify_fg(STATE_NORMAL, label_color)
+            if label_font:
+                label.modify_font(label_font)
     return button
 
 def menuitem(title, image=None, tooltip=None, cb=None):
@@ -1166,7 +1436,7 @@ def window_defaults(window):
 #selected in a submenu:
 def ensure_item_selected(submenu, item, recurse=True):
     if not isinstance(item, gtk.CheckMenuItem):
-        return
+        return None
     if item.get_active():
         #deactivate all except this one
         def deactivate(items, skip=None):
@@ -1197,7 +1467,7 @@ def ensure_item_selected(submenu, item, recurse=True):
         return None
     active = get_active_item(submenu.get_children())
     if active:
-        return  active
+        return active
     #if not then keep this one active:
     item.set_active(True)
     return item
@@ -1231,7 +1501,8 @@ class TableBuilder(object):
         self.inc()
 
     def attach(self, widget, i, count=1, xoptions=FILL, yoptions=FILL, xpadding=10, ypadding=0):
-        self.table.attach(widget, i, i+count, self.row, self.row+1, xoptions=xoptions, yoptions=yoptions, xpadding=xpadding, ypadding=ypadding)
+        self.table.attach(widget, i, i+count, self.row, self.row+1,
+                          xoptions=xoptions, yoptions=yoptions, xpadding=xpadding, ypadding=ypadding)
 
     def inc(self):
         self.row += 1
@@ -1241,7 +1512,7 @@ class TableBuilder(object):
         self.add_row(row_label, value1, value2, **kwargs)
 
 
-def choose_file(parent_window, title, action, action_button, callback, file_filter=None):
+def choose_file(parent_window, title, action, action_button, callback=None, file_filter=None):
     log("choose_file%s", (parent_window, title, action, action_button, callback, file_filter))
     chooser = gtk.FileChooserDialog(title,
                                 parent=parent_window, action=action,
@@ -1255,10 +1526,11 @@ def choose_file(parent_window, title, action, action_button, callback, file_filt
     chooser.hide()
     chooser.destroy()
     if response!=RESPONSE_OK or len(filenames)!=1:
-        return
+        return None
     filename = filenames[0]
-    callback(filename)
-
+    if callback:
+        callback(filename)
+    return filename
 
 
 def main():
