@@ -30,7 +30,7 @@ def parse_uid(v):
     if v:
         try:
             return int(v)
-        except:
+        except (TypeError, ValueError):
             log("uid '%s' is not an int", v)
     if POSIX:
         try:
@@ -46,7 +46,7 @@ def parse_gid(v):
     if v:
         try:
             return int(v)
-        except:
+        except (TypeError, ValueError):
             log("gid '%s' is not an int", v)
     if POSIX:
         try:
@@ -96,7 +96,7 @@ class SysAuthenticatorBase(object):
         return self.salt, self.digest
 
     def get_passwords(self):
-        p = self.get_password()
+        p = self.get_password()     #pylint: disable=assignment-from-none
         if p is not None:
             return (p,)
         return ()
@@ -145,6 +145,7 @@ class SysAuthenticatorBase(object):
             ret = self.check(password)
             log("authenticate_check(..)=%s", ret)
         except Exception as e:
+            log("check(..)", exc_info=True)
             log.error("Error: %s authentication check failed:", self)
             log.error(" %s", e)
             return False
@@ -196,12 +197,13 @@ class SysAuthenticator(SysAuthenticatorBase):
 
     def __init__(self, username, **kwargs):
         SysAuthenticatorBase.__init__(self, username)
-        try:
-            import pwd
-            self.pw = pwd.getpwnam(username)
-        except Exception:
-            log("cannot load password database entry for '%s'", username, exc_info=True)
-            self.pw = None
+        self.pw = None
+        if POSIX:
+            try:
+                import pwd
+                self.pw = pwd.getpwnam(username)
+            except Exception:
+                log("cannot load password database entry for '%s'", username, exc_info=True)
 
     def get_uid(self):
         if self.pw is None:
