@@ -80,7 +80,7 @@ class InputServer(StubServerMixin):
     def parse_hello_ui_keyboard(self, ss, c):
         other_ui_clients = [s.uuid for s in self._server_sources.values() if s!=ss and s.ui_client]
         #parse client config:
-        ss.keyboard_config = self.get_keyboard_config(c)
+        ss.keyboard_config = self.get_keyboard_config(c)    #pylint: disable=assignment-from-none
 
         if not other_ui_clients:
             #so only activate this feature afterwards:
@@ -119,7 +119,7 @@ class InputServer(StubServerMixin):
             options = packet[3]
         else:
             options = ""
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss and ss.set_layout(layout, variant, options):
             self.set_keymap(ss, force=True)
 
@@ -127,7 +127,7 @@ class InputServer(StubServerMixin):
         if self.readonly:
             return
         props = typedict(packet[1])
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         keylog("received new keymap from client")
@@ -150,7 +150,7 @@ class InputServer(StubServerMixin):
         if self.readonly:
             return
         wid, keyname, pressed, modifiers, keyval, _, client_keycode, group = packet[1:9]
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         keyname = bytestostr(keyname)
@@ -248,7 +248,7 @@ class InputServer(StubServerMixin):
         if self.readonly:
             return
         wid, keyname, keyval, client_keycode, modifiers = packet[1:6]
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         keyname = bytestostr(keyname)
@@ -279,7 +279,7 @@ class InputServer(StubServerMixin):
         assert proto in self._server_sources
         if self.readonly:
             return
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         ss.keyboard_config.sync = bool(packet[1])
@@ -310,7 +310,7 @@ class InputServer(StubServerMixin):
 
     def _adjust_pointer(self, proto, wid, pointer):
         #the window may not be mapped at the same location by the client:
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         window = self._id_to_window.get(wid)
         if ss and window:
             ws = ss.get_window_source(wid)
@@ -346,7 +346,7 @@ class InputServer(StubServerMixin):
         mouselog("process_button_action(%s, %s)", proto, packet)
         if self.readonly:
             return
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         ss.user_event()
@@ -361,10 +361,11 @@ class InputServer(StubServerMixin):
         pass
 
     def _process_pointer_position(self, proto, packet):
-        mouselog("_process_pointer_position(%s, %s) readonly=%s, ui_driver=%s", proto, packet, self.readonly, self.ui_driver)
+        mouselog("_process_pointer_position(%s, %s) readonly=%s, ui_driver=%s",
+                 proto, packet, self.readonly, self.ui_driver)
         if self.readonly:
             return
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         wid, pdata, modifiers = packet[1:4]
@@ -393,26 +394,26 @@ class InputServer(StubServerMixin):
         pass
 
 
-    def send_hello(self, server_source, root_w, root_h, key_repeat, server_cipher):
+    def send_hello(self, server_source, _root_w, _root_h, key_repeat, _server_cipher):
         capabilities = self.make_hello(server_source)
         if key_repeat:
             capabilities.update({
-                     "key_repeat"           : key_repeat,
-                     "key_repeat_modifiers" : True})
+                "key_repeat"           : key_repeat,
+                "key_repeat_modifiers" : True,
+                })
 
 
     def init_packet_handlers(self):
-        self._authenticated_packet_handlers.update({
-            "set-keyboard-sync-enabled":            self._process_keyboard_sync_enabled_status,
-          })
-        self._authenticated_ui_packet_handlers.update({
-            "key-action":                           self._process_key_action,
-            "key-repeat":                           self._process_key_repeat,
-            "layout-changed":                       self._process_layout,
-            "keymap-changed":                       self._process_keymap,
+        self.add_packet_handlers({
+            #keyboard:
+            "set-keyboard-sync-enabled" : self._process_keyboard_sync_enabled_status,
+            "key-action"                : self._process_key_action,
+            "key-repeat"                : self._process_key_repeat,
+            "layout-changed"            : self._process_layout,
+            "keymap-changed"            : self._process_keymap,
             #mouse:
-            "button-action":                        self._process_button_action,
-            "pointer-position":                     self._process_pointer_position,
+            "button-action"             : self._process_button_action,
+            "pointer-position"          : self._process_pointer_position,
             #setup:
-            "input-devices":                        self._process_input_devices,
+            "input-devices"             : self._process_input_devices,
             })
