@@ -1,45 +1,39 @@
 # -*- coding: utf-8 -*-
 # This file is part of Xpra.
 # Copyright (C) 2011 Serviware (Arthur Huillet, <ahuillet@serviware.com>)
-# Copyright (C) 2010-2017 Antoine Martin <antoine@devloop.org.uk>
+# Copyright (C) 2010-2018 Antoine Martin <antoine@xpra.org>
 # Copyright (C) 2008 Nathaniel Smith <njs@pobox.com>
 # Xpra is released under the terms of the GNU GPL v2, or, at your option, any
 # later version. See the file COPYING for details.
 
 import os
-import gtk
-from gtk import gdk
-import glib
-import gobject
 import signal
 import math
 from collections import deque, namedtuple
 
 from xpra.version_util import XPRA_VERSION
-from xpra.util import updict, rindex, envbool, envint
-from xpra.os_util import memoryview_to_bytes, monotonic_time
+from xpra.util import updict, rindex, envbool, envint, typedict
+from xpra.os_util import memoryview_to_bytes, strtobytes, bytestostr, monotonic_time
 from xpra.server import server_features
 from xpra.gtk_common.gobject_util import one_arg_signal
-from xpra.gtk_common.gtk_util import get_default_root_window, get_xwindow
-from xpra.x11.common import Unmanageable
+from xpra.gtk_common.gtk_util import (
+    get_default_root_window, get_xwindow, pixbuf_new_from_data, is_realized,
+    SUBSTRUCTURE_MASK, GDKWINDOW_TEMP,
+    )
+from xpra.x11.common import Unmanageable, MAX_WINDOW_SIZE
 from xpra.x11.gtk_x11.prop import prop_set
-from xpra.x11.gtk2.wm import Wm
-from xpra.x11.gtk2.tray import get_tray_window, SystemTray
-from xpra.x11.gtk2.window import OverrideRedirectWindowModel, SystemTrayWindowModel
-from xpra.x11.gtk2.gdk_bindings import (
-                               add_event_receiver,          #@UnresolvedImport
-                               get_children,                #@UnresolvedImport
-                               init_x11_filter,             #@UnresolvedImport
-                               cleanup_x11_filter,          #@UnresolvedImport
-                               cleanup_all_event_receivers  #@UnresolvedImport
-                               )
+from xpra.x11.gtk_x11.tray import get_tray_window, SystemTray
+from xpra.x11.gtk_x11.gdk_bindings import (
+   add_event_receiver,
+   get_children,
+   )
 from xpra.x11.bindings.window_bindings import X11WindowBindings #@UnresolvedImport
-X11Window = X11WindowBindings()
 from xpra.x11.bindings.keyboard_bindings import X11KeyboardBindings #@UnresolvedImport
-X11Keyboard = X11KeyboardBindings()
-from xpra.gtk_common.error import xsync, xswallow
-
+from xpra.x11.x11_server_base import X11ServerBase
+from xpra.gtk_common.error import xsync, xswallow, xlog, XError
+from xpra.gtk_common.gobject_compat import import_gtk, import_gdk, import_glib, import_gobject, is_gtk3
 from xpra.log import Logger
+
 log = Logger("server")
 focuslog = Logger("server", "focus")
 grablog = Logger("server", "grab")
@@ -49,16 +43,22 @@ traylog = Logger("server", "tray")
 workspacelog = Logger("x11", "workspace")
 metadatalog = Logger("x11", "metadata")
 framelog = Logger("x11", "frame")
-menulog  = Logger("x11", "menu")
 eventlog = Logger("x11", "events")
 mouselog = Logger("x11", "mouse")
 
-from xpra.x11.x11_server_base import X11ServerBase
+X11Window = X11WindowBindings()
+X11Keyboard = X11KeyboardBindings()
+
+gtk = import_gtk()
+gdk = import_gdk()
+glib = import_glib()
+gobject = import_gobject()
 
 REPARENT_ROOT = envbool("XPRA_REPARENT_ROOT", True)
 CONFIGURE_DAMAGE_RATE = envint("XPRA_CONFIGURE_DAMAGE_RATE", 250)
 SHARING_SYNC_SIZE = envbool("XPRA_SHARING_SYNC_SIZE", True)
 CLAMP_WINDOW_TO_ROOT = envbool("XPRA_CLAMP_WINDOW_TO_ROOT", False)
+ALWAYS_RAISE_WINDOW = envbool("XPRA_ALWAYS_RAISE_WINDOW", False)
 
 WINDOW_SIGNALS = os.environ.get("XPRA_WINDOW_SIGNALS", "SIGINT,SIGTERM,SIGQUIT,SIGCONT,SIGUSR1,SIGUSR2").split(",")
 
@@ -76,7 +76,10 @@ class DesktopManager(gtk.Widget):
         self._models = {}
         gtk.Widget.__init__(self)
         self.set_property("can-focus", True)
-        self.set_flags(gtk.NO_WINDOW)
+        if is_gtk3():
+            self.realize()
+        else:
+            self.set_flags(gtk.NO_WINDOW)
 
     def __repr__(self):
         return "DesktopManager(%s)" % len(self._models)
@@ -84,7 +87,8 @@ class DesktopManager(gtk.Widget):
     ## For communicating with the main WM:
 
     def add_window(self, model, x, y, w, h):
-        assert self.flags() & gtk.REALIZED
+        if not is_gtk3():
+            assert is_realized(self)
         self._models[model] = DesktopState([x, y, w, h])
         model.managed_connect("unmanaged", self._unmanaged)
         model.managed_connect("ownership-election", self._elect_me)
@@ -153,14 +157,21 @@ class DesktopManager(gtk.Widget):
     def _elect_me(self, model):
         if self.visible(model):
             return (1, self)
-        else:
-            return (-1, self)
+        return (-1, self)
 
     def take_window(self, _model, window):
+        #log.info("take_window(%s, %s)", model, window)
+        if not is_realized(self):
+            assert is_gtk3()
+            #with GTK3, the widget is never realized??
+            return
+        gdkwin = self.get_window()
+        assert gdkwin, "DesktopManager does not have a gdk window"
         if REPARENT_ROOT:
-            parent = self.window.get_screen().get_root_window()
+            parent = gdkwin.get_screen().get_root_window()
         else:
-            parent = self.window
+            parent = gdkwin
+        log("take_window: reparenting %s to %s", window, parent)
         window.reparent(parent, 0, 0)
 
     def window_size(self, model):
@@ -168,7 +179,7 @@ class DesktopManager(gtk.Widget):
         return w, h
 
     def window_position(self, model, w=None, h=None):
-        [x, y, w0, h0] = self._models[model].geom
+        x, y, w0, h0 = self._models[model].geom
         if (w is not None and abs(w0-w)>1) or (h is not None and abs(h0-h)>1):
             log("Uh-oh, our size doesn't fit window sizing constraints: "
                      "%sx%s vs %sx%s", w0, h0, w, h)
@@ -185,16 +196,17 @@ class XpraServer(gobject.GObject, X11ServerBase):
         }
 
     def __init__(self, clobber):
-        self.clobber = clobber
         self.root_overlay = None
         self.repaint_root_overlay_timer = None
         self.configure_damage_timers = {}
         self._tray = None
+        self._has_grab = 0
+        self._has_focus = 0
+        self._wm = None
         self.last_raised = None
         self.system_tray = False
-        self.global_menus = False
         gobject.GObject.__init__(self)
-        X11ServerBase.__init__(self)
+        X11ServerBase.__init__(self, clobber)
         self.session_type = "seamless"
 
     def init(self, opts):
@@ -203,16 +215,28 @@ class XpraServer(gobject.GObject, X11ServerBase):
         self.system_tray = opts.system_tray
         X11ServerBase.init(self, opts)
         self.fake_xinerama = opts.fake_xinerama
-        #rpc_handlers are part of DBUS_RPC_Server, which is optional
-        rpc_handlers = getattr(self, "rpc_handlers", None)
-        if rpc_handlers:
-            self.global_menus = int(opts.global_menus)
-            if self.global_menus:
-                self.rpc_handlers["menu"] = self._handle_menu_rpc
         def log_server_event(_, event):
             eventlog("server-event: %s", event)
         self.connect("server-event", log_server_event)
 
+    def server_init(self):
+        X11ServerBase.server_init(self)
+        if not self.clobber and self.randr:
+            from xpra.x11.vfb_util import set_initial_resolution, DEFAULT_VFB_RESOLUTION
+            with xlog:
+                set_initial_resolution(DEFAULT_VFB_RESOLUTION)
+
+    def server_ready(self):
+        if not X11Window.displayHasXComposite():
+            log.error("Xpra 'start' subcommand runs as a compositing manager")
+            log.error(" it cannot use a display which lacks the XComposite extension!")
+            return False
+        #check for an existing window manager:
+        from xpra.x11.gtk_x11.wm_check import wm_check
+        if not wm_check(self.wm_name, self.clobber):
+            return False
+        return True
+
     def setup(self):
         X11ServerBase.setup(self)
         if self.system_tray:
@@ -220,25 +244,19 @@ class XpraServer(gobject.GObject, X11ServerBase):
 
     def x11_init(self):
         X11ServerBase.x11_init(self)
-        assert init_x11_filter() is True
 
-        self._has_grab = 0
-        self._has_focus = 0
         self._focus_history = deque(maxlen=100)
         # Do this before creating the Wm object, to avoid clobbering its
         # selecting SubstructureRedirect.
-        root = gdk.get_default_root_window()
-        root.set_events(root.get_events() | gdk.SUBSTRUCTURE_MASK)
-        root.property_change(gdk.atom_intern("XPRA_SERVER", False),
-                            gdk.atom_intern("STRING", False),
-                            8,
-                            gdk.PROP_MODE_REPLACE,
-                            XPRA_VERSION)
+        root = get_default_root_window()
+        root.set_events(root.get_events() | SUBSTRUCTURE_MASK)
+        prop_set(root, "XPRA_SERVER", "latin1", strtobytes(XPRA_VERSION).decode())
         add_event_receiver(root, self)
         if self.sync_xvfb>0:
+            xid = get_xwindow(root)
             try:
                 with xsync:
-                    self.root_overlay = X11Window.XCompositeGetOverlayWindow(root.xid)
+                    self.root_overlay = X11Window.XCompositeGetOverlayWindow(xid)
                     if self.root_overlay:
                         #ugly: API expects a window object with a ".xid"
                         X11WindowModel = namedtuple("X11WindowModel", "xid")
@@ -246,7 +264,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
                         prop_set(root_overlay, "WM_TITLE", "latin1", u"RootOverlay")
                         X11Window.AllowInputPassthrough(self.root_overlay)
             except Exception as e:
-                log("XCompositeGetOverlayWindow(%#x)", root.xid, exc_info=True)
+                log("XCompositeGetOverlayWindow(%#x)", xid, exc_info=True)
                 log.error("Error setting up xvfb synchronization:")
                 log.error(" %s", e)
                 if self.root_overlay:
@@ -255,6 +273,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
                     self.root_overlay = None
 
         ### Create the WM object
+        from xpra.x11.gtk_x11.wm import Wm
         self._wm = Wm(self.clobber, self.wm_name)
         if server_features.windows:
             self._wm.connect("new-window", self._new_window_signaled)
@@ -275,9 +294,6 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 X11Window.XCompositeReleaseOverlayWindow(self.root_overlay)
             self.root_overlay = None
         self.cancel_all_configure_damage()
-        X11ServerBase.do_cleanup(self)
-        cleanup_x11_filter()
-        cleanup_all_event_receivers()
         if self._wm:
             self._wm.cleanup()
             self._wm = None
@@ -286,6 +302,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
             #but at this point in the cleanup, we probably won't, so force set it:
             self._has_grab = 0
             self.X11_ungrab()
+        X11ServerBase.do_cleanup(self)
 
 
     def last_client_exited(self):
@@ -296,15 +313,26 @@ class XpraServer(gobject.GObject, X11ServerBase):
             self.X11_ungrab()
 
 
+    def update_size_constraints(self, minw, minh, maxw, maxh):
+        wm = self._wm
+        if wm:
+            wm.set_size_constraints(minw, minh, maxw, maxh)
+        elif server_features.windows:
+            #update the default so Wm will use it
+            #when we do intantiate it:
+            from xpra.x11.gtk_x11 import wm
+            wm.DEFAULT_SIZE_CONSTRAINTS = (0, 0, MAX_WINDOW_SIZE, MAX_WINDOW_SIZE)
+
+
     def init_packet_handlers(self):
         X11ServerBase.init_packet_handlers(self)
-        self._authenticated_ui_packet_handlers.update({
-            "window-signal"         : self._process_window_signal,
-            })
+        self.add_packet_handler("window-signal", self._process_window_signal)
 
 
     def get_server_mode(self):
-        return "X11"
+        if is_gtk3():
+            return "GTK3 X11"
+        return "GTK2 X11"
 
 
     def server_event(self, *args):
@@ -326,7 +354,11 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 "configure.delta"        : True,
                 "signals"                : WINDOW_SIGNALS,
                 "dragndrop"              : True,
-                "states"                 : ["iconified", "fullscreen", "above", "below", "sticky", "iconified", "maximized"],
+                "states"                 : [
+                    "iconified", "focused", "fullscreen",
+                    "above", "below",
+                    "sticky", "iconified", "maximized",
+                    ],
                 })
         return capabilities
 
@@ -372,7 +404,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         #only run the default code if there are no clients,
         #when we have clients, this should have been done already
         #in the code that synchonizes the screen resolution
-        if len(self._server_sources)==0:
+        if not self._server_sources:
             super(XpraServer, self).set_screen_geometry_attributes(w, h)
 
     def set_desktops(self, names):
@@ -413,17 +445,22 @@ class XpraServer(gobject.GObject, X11ServerBase):
         return self._desktop_manager.is_shown(window)
 
     def load_existing_windows(self):
+        if not self._wm:
+            return
 
         ### Create our window managing data structures:
         self._desktop_manager = DesktopManager()
-        self._wm.get_property("toplevel").add(self._desktop_manager)
+        add_to = self._wm.get_property("toplevel")
+        #may be missing with GTK3..
+        if add_to:
+            add_to.add(self._desktop_manager)
         self._desktop_manager.show_all()
 
         ### Load in existing windows:
         for window in self._wm.get_property("windows"):
             self._add_new_window(window)
 
-        root = gdk.get_default_root_window()
+        root = get_default_root_window()
         with xsync:
             for window in get_children(root):
                 xid = get_xwindow(window)
@@ -438,7 +475,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         transient_for = window.get_property("transient-for")
         log("get_transient_for window=%s, transient_for=%s", window, transient_for)
         if transient_for is None:
-            return None
+            return 0
         xid = get_xwindow(transient_for)
         log("transient_for.xid=%#x", xid)
         for w,wid in self._window_to_id.items():
@@ -450,7 +487,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
             log("transient-for using root")
             return -1       #-1 is the backwards compatible marker for root...
         log("not found transient_for=%s, xid=%#x", transient_for, xid)
-        return  None
+        return 0
 
 
     def parse_hello_ui_window_settings(self, ss, _caps):
@@ -458,7 +495,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
         frame = None
         if ss.window_frame_sizes:
             frame = ss.window_frame_sizes.intlistget("frame", (0, 0, 0, 0), 4, 4)
-        self._wm.set_default_frame_extents(frame)
+        if self._wm:
+            self._wm.set_default_frame_extents(frame)
 
 
     def send_initial_windows(self, ss, sharing=False):
@@ -518,6 +556,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         window.managed_connect("ungrab", self._window_ungrab)
         window.managed_connect("bell", self._bell_signaled)
         window.managed_connect("motion", self._motion_signaled)
+        window.managed_connect("x11-property-changed", self._x11_property_changed)
         if not window.is_tray():
             window.managed_connect("raised", self._raised_window)
             window.managed_connect("initiate-moveresize", self._initiate_moveresize)
@@ -530,6 +569,17 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 #but we don't really care, it will get cleaned up soon enough
         return wid
 
+
+    def _x11_property_changed(self, window, event):
+        #name, dtype, dformat, value = event
+        metadata = {"x11-property" : event}
+        wid = self._window_to_id[window]
+        for ss in self._server_sources.values():
+            ms = getattr(ss, "metadata_supported", ())
+            if "x11-property" in ms:
+                ss.send("window-metadata", wid, metadata)
+
+
     def _add_new_window(self, window):
         self._add_new_window_common(window)
         x, y, w, h = window.get_property("geometry")
@@ -542,7 +592,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
     def _window_resized_signaled(self, window, *args):
         x, y, nw, nh = window.get_property("geometry")[:4]
         geom = self._desktop_manager.window_geometry(window)
-        geomlog("XpraServer._window_resized_signaled(%s,%s) geometry=%s, desktop manager geometry=%s", window, args, (x, y, nw, nh), geom)
+        geomlog("XpraServer._window_resized_signaled(%s,%s) geometry=%s, desktop manager geometry=%s",
+                window, args, (x, y, nw, nh), geom)
         if geom==[x, y, nw, nh]:
             geomlog("XpraServer._window_resized_signaled: unchanged")
             #unchanged
@@ -561,7 +612,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
         self.snc_timer = glib.timeout_add(int(delay), self.size_notify_clients, window, lcce)
 
     def size_notify_clients(self, window, lcce=-1):
-        geomlog("size_notify_clients(%s, %s) last_client_configure_event=%s", window, lcce, self.last_client_configure_event)
+        geomlog("size_notify_clients(%s, %s) last_client_configure_event=%s",
+                window, lcce, self.last_client_configure_event)
         self.snc_timer = 0
         wid = self._window_to_id.get(window)
         if not wid:
@@ -584,19 +636,19 @@ class XpraServer(gobject.GObject, X11ServerBase):
         if self.root_overlay and self.root_overlay==xid:
             windowlog("ignoring root overlay window %#x", self.root_overlay)
             return
-        if raw_window.get_window_type()==gdk.WINDOW_TEMP:
+        if raw_window.get_window_type()==GDKWINDOW_TEMP:
             #ignoring one of gtk's temporary windows
             #all the windows we manage should be gdk.WINDOW_FOREIGN
             windowlog("ignoring TEMP window %#x", xid)
             return
         WINDOW_MODEL_KEY = "_xpra_window_model_"
-        wid = raw_window.get_data(WINDOW_MODEL_KEY)
+        wid = getattr(raw_window, WINDOW_MODEL_KEY, None)
         window = self._id_to_window.get(wid)
         if window:
             if window.is_managed():
                 windowlog("found existing window model %s for %#x, will refresh it", type(window), xid)
                 ww, wh = window.get_dimensions()
-                self._damage(window, 0, 0, ww, wh, options={"min_delay" : 50})
+                self.refresh_window_area(window, 0, 0, ww, wh, options={"min_delay" : 50})
                 return
             windowlog("found existing model %s (but no longer managed!) for %#x", type(window), xid)
             #we could try to re-use the existing model and window ID,
@@ -613,15 +665,17 @@ class XpraServer(gobject.GObject, X11ServerBase):
             tray_window = get_tray_window(raw_window)
             if tray_window is not None:
                 assert self._tray
+                from xpra.x11.models.systray import SystemTrayWindowModel
                 window = SystemTrayWindowModel(raw_window)
                 wid = self._add_new_window_common(window)
-                raw_window.set_data(WINDOW_MODEL_KEY, wid)
+                setattr(raw_window, WINDOW_MODEL_KEY, wid)
                 window.call_setup()
                 self._send_new_tray_window_packet(wid, window)
             else:
+                from xpra.x11.models.or_window import OverrideRedirectWindowModel
                 window = OverrideRedirectWindowModel(raw_window)
                 wid = self._add_new_window_common(window)
-                raw_window.set_data(WINDOW_MODEL_KEY, wid)
+                setattr(raw_window, WINDOW_MODEL_KEY, wid)
                 window.call_setup()
                 window.managed_connect("notify::geometry", self._or_window_geometry_changed)
                 self._send_new_or_window_packet(window)
@@ -707,7 +761,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
             focuslog.warn("Warning: window %s is no longer managed!", window)
             return
         focuslog("focus: giving focus to %s", window)
-        window.raise_window()
+        with xswallow:
+            window.raise_window()
         window.give_client_focus()
         if server_source and modifiers is not None:
             make_keymask_match = getattr(server_source, "make_keymask_match", None)
@@ -727,14 +782,13 @@ class XpraServer(gobject.GObject, X11ServerBase):
     def _send_new_or_window_packet(self, window):
         geometry = window.get_property("geometry")
         self._do_send_new_window_packet("new-override-redirect", window, geometry)
-        ww, wh = window.get_dimensions()
-        self._damage(window, 0, 0, ww, wh)
+        self.refresh_window(window)
 
     def _send_new_tray_window_packet(self, wid, window):
         ww, wh = window.get_dimensions()
         for ss in self._server_sources.values():
             ss.new_tray(wid, window, ww, wh)
-        self._damage(window, 0, 0, ww, wh)
+        self.refresh_window(window)
 
 
     def _lost_window(self, window, wm_exiting=False):
@@ -745,12 +799,13 @@ class XpraServer(gobject.GObject, X11ServerBase):
 
     def _contents_changed(self, window, event):
         if window.is_OR() or window.is_tray() or self._desktop_manager.visible(window):
-            self._damage(window, event.x, event.y, event.width, event.height)
+            self.refresh_window_area(window, event.x, event.y, event.width, event.height, options={"damage" : True})
 
 
     def _window_grab(self, window, event):
         grab_id = self._window_to_id.get(window, -1)
-        grablog("window_grab(%s, %s) has_grab=%s, has focus=%s, grab window=%s", window, event, self._has_grab, self._has_focus, grab_id)
+        grablog("window_grab(%s, %s) has_grab=%s, has focus=%s, grab window=%s",
+                window, event, self._has_grab, self._has_focus, grab_id)
         if grab_id<0 or self._has_grab==grab_id:
             return
         self._has_grab = grab_id
@@ -759,7 +814,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
 
     def _window_ungrab(self, window, event):
         grab_id = self._window_to_id.get(window, -1)
-        grablog("window_ungrab(%s, %s) has_grab=%s, has focus=%s, grab window=%s", window, event, self._has_grab, self._has_focus, grab_id)
+        grablog("window_ungrab(%s, %s) has_grab=%s, has focus=%s, grab window=%s",
+                window, event, self._has_grab, self._has_focus, grab_id)
         if not self._has_grab:
             return
         self._has_grab = 0
@@ -791,14 +847,15 @@ class XpraServer(gobject.GObject, X11ServerBase):
         assert proto in self._server_sources
         if not new_window_state:
             return []
+        nws = typedict(new_window_state)
         metadatalog("set_window_state%s", (wid, window, new_window_state))
         changes = []
         if "frame" in new_window_state:
             #the size of the window frame may have changed
-            frame = new_window_state.get("frame") or (0, 0, 0, 0)
+            frame = nws.intlistget("frame", (0, 0, 0, 0))
             window.set_property("frame", frame)
         #boolean: but not a wm_state and renamed in the model... (iconic vs inconified!)
-        iconified = new_window_state.get("iconified")
+        iconified = nws.boolget("iconified")
         if iconified is not None:
             if window.is_OR():
                 log("ignoring iconified=%s on OR window %s", iconified, window)
@@ -807,15 +864,21 @@ class XpraServer(gobject.GObject, X11ServerBase):
                     window.set_property("iconic", iconified)
                     changes.append("iconified")
         #handle wm_state virtual booleans:
-        for k in ("maximized", "above", "below", "fullscreen", "sticky", "shaded", "skip-pager", "skip-taskbar", "focused"):
-            if k in new_window_state:
-                #metadatalog.info("window.get_property=%s", window.get_property)
-                new_state = bool(new_window_state.get(k, False))
-                cur_state = bool(window.get_property(k))
-                #metadatalog.info("set window state for '%s': current state=%s, new state=%s", k, cur_state, new_state)
-                if cur_state!=new_state:
-                    window.update_wm_state(k, new_state)
-                    changes.append(k)
+        for k in (
+            "maximized", "above",
+            "below", "fullscreen",
+            "sticky", "shaded",
+            "skip-pager", "skip-taskbar", "focused",
+            ):
+            #metadatalog.info("window.get_property=%s", window.get_property)
+            new_state = nws.boolget(k, None)
+            if new_state is None:
+                continue
+            cur_state = bool(window.get_property(k))
+            #metadatalog.info("set window state for '%s': current state=%s, new state=%s", k, cur_state, new_state)
+            if cur_state!=new_state:
+                window.update_wm_state(k, new_state)
+                changes.append(k)
         metadatalog("set_window_state: changes=%s", changes)
         return changes
 
@@ -834,22 +897,22 @@ class XpraServer(gobject.GObject, X11ServerBase):
             windowlog("cannot map window %i: not found, already removed?", wid)
             return
         assert not window.is_OR()
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         geomlog("client %s mapped window %i - %s, at: %s", ss, wid, window, (x, y, w, h))
         self._window_mapped_at(proto, wid, window, (x, y, w, h))
         if len(packet)>=7:
             self._set_client_properties(proto, wid, window, packet[6])
-        #if not self.ui_driver:
-        #    self.set_ui_driver(ss)
+        if not self.ui_driver:
+            self.set_ui_driver(ss)
         if self.ui_driver==ss.uuid or not self._desktop_manager.is_shown(window):
             if len(packet)>=8:
                 self._set_window_state(proto, wid, window, packet[7])
             ax, ay, aw, ah = self._clamp_window(proto, wid, window, x, y, w, h)
             self._desktop_manager.configure_window(window, ax, ay, aw, ah)
             self._desktop_manager.show_window(window)
-        self._damage(window, 0, 0, w, h)
+        self.refresh_window_area(window, 0, 0, w, h)
 
 
     def _process_unmap_window(self, proto, packet):
@@ -859,7 +922,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
             log("cannot unmap window %i: not found, already removed?", wid)
             return
         assert not window.is_OR()
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if ss is None:
             return
         self._window_mapped_at(proto, wid, window)
@@ -890,7 +953,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
             x = max(0, min(x, rw-w))
             y = max(0, min(y, rh-h))
             #tell this client to honour the new location
-            ss = self._server_sources.get(proto)
+            ss = self.get_server_source(proto)
             if ss:
                 resize_counter = max(resize_counter, self._desktop_manager.get_resize_counter(window, 1))
                 ss.move_resize_window(wid, window, x, y, w, h, resize_counter)
@@ -902,7 +965,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         if not window:
             geomlog("cannot configure window %i: not found, already removed?", wid)
             return
-        ss = self._server_sources.get(proto)
+        ss = self.get_server_source(proto)
         if not ss:
             return
         #some "configure-window" packets are only meant for metadata updates:
@@ -914,11 +977,11 @@ class XpraServer(gobject.GObject, X11ServerBase):
             if cprops:
                 metadatalog("window client properties updates: %s", cprops)
                 self._set_client_properties(proto, wid, window, cprops)
-        #if not self.ui_driver:
-        #    self.set_ui_driver(ss)
+        if not self.ui_driver:
+            self.set_ui_driver(ss)
         is_ui_driver = self.ui_driver==ss.uuid
         shown = self._desktop_manager.is_shown(window)
-        if window.is_OR() or window.is_tray() or skip_geometry:
+        if window.is_OR() or window.is_tray() or skip_geometry or self.readonly:
             size_changed = False
         else:
             oww, owh = self._desktop_manager.window_size(window)
@@ -935,7 +998,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
                         self._update_modifiers(proto, wid, modifiers)
             if window.is_tray():
                 assert self._tray
-                if not skip_geometry:
+                if not skip_geometry and not self.readonly:
                     traylog("tray %s configured to: %s", window, (x, y, w, h))
                     self._tray.move_resize(window, x, y, w, h)
                     damage = True
@@ -944,7 +1007,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 self.last_client_configure_event = monotonic_time()
                 if is_ui_driver and len(packet)>=9:
                     changes = self._set_window_state(proto, wid, window, packet[8])
-                    damage |= len(changes)>0
+                    if changes:
+                        damage = True
                 if not skip_geometry:
                     owx, owy, oww, owh = self._desktop_manager.window_geometry(window)
                     resize_counter = 0
@@ -978,12 +1042,11 @@ class XpraServer(gobject.GObject, X11ServerBase):
         def damage():
             try:
                 del self.configure_damage_timers[wid]
-            except:
+            except KeyError:
                 pass
             window = self._lookup_window(wid)
             if window and window.is_managed():
-                w, h = window.get_dimensions()
-                self._damage(window, 0, 0, w, h)
+                self.refresh_window(window)
         self.configure_damage_timers[wid] = self.timeout_add(CONFIGURE_DAMAGE_RATE, damage)
 
     def cancel_configure_damage(self, wid):
@@ -991,7 +1054,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         if timer:
             try:
                 del self.configure_damage_timers[wid]
-            except:
+            except KeyError:
                 pass
             self.source_remove(timer)
 
@@ -1020,14 +1083,15 @@ class XpraServer(gobject.GObject, X11ServerBase):
     """ override so we can raise the window under the cursor
         (gtk raise does not change window stacking, just focus) """
     def _move_pointer(self, wid, pos, *args):
-        if wid>0 and self.last_raised!=wid:
+        if wid>0 and (self.last_raised!=wid or ALWAYS_RAISE_WINDOW):
             window = self._lookup_window(wid)
             if not window:
                 mouselog("_move_pointer(%s, %s) invalid window id", wid, pos)
             else:
                 self.last_raised = wid
                 mouselog("raising %s", window)
-                window.raise_window()
+                with xswallow:
+                    window.raise_window()
         X11ServerBase._move_pointer(self, wid, pos, *args)
 
 
@@ -1046,7 +1110,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
     def _process_window_signal(self, proto, packet):
         assert proto in self._server_sources
         wid = packet[1]
-        sig = packet[2]
+        sig = bytestostr(packet[2])
         if sig not in WINDOW_SIGNALS:
             log.warn("Warning: window signal '%s' not handled", sig)
             return
@@ -1069,8 +1133,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
             log.error(" %s", e)
 
 
-    def _damage(self, window, x, y, width, height, options=None):
-        super(XpraServer, self)._damage(window, x, y, width, height, options)
+    def refresh_window_area(self, window, x, y, width, height, options=None):
+        super(XpraServer, self).refresh_window_area(window, x, y, width, height, options)
         if self.root_overlay:
             image = window.get_image(x, y, width, height)
             if image:
@@ -1106,7 +1170,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
         log("update_root_overlay%s painting rectangle %s", (window, x, y, image), (wx+x, wy+y, width, height))
         if has_alpha:
             import cairo
-            pixbuf = gdk.pixbuf_new_from_data(img_data, gdk.COLORSPACE_RGB, True, 8, width, height, rowstride)
+            pixbuf = pixbuf_new_from_data(img_data, gdk.COLORSPACE_RGB, True, 8, width, height, rowstride)
             cr = overlaywin.cairo_create()
             cr.new_path()
             cr.rectangle(wx+x, wy+y, width, height)
@@ -1119,8 +1183,11 @@ class XpraServer(gobject.GObject, X11ServerBase):
         image.free()
 
     def repaint_root_overlay(self):
-        log("repaint_root_overlay() root_overlay=%s, due=%s, sync-xvfb=%ims", self.root_overlay, self.repaint_root_overlay_timer, self.sync_xvfb)
-        if not self.root_overlay or self.repaint_root_overlay_timer:
+        if not self.root_overlay:
+            return
+        log("repaint_root_overlay() root_overlay=%s, due=%s, sync-xvfb=%ims",
+            self.root_overlay, self.repaint_root_overlay_timer, self.sync_xvfb)
+        if self.repaint_root_overlay_timer:
             return
         self.repaint_root_overlay_timer = self.timeout_add(self.sync_xvfb, self.do_repaint_root_overlay)
 
@@ -1187,7 +1254,8 @@ class XpraServer(gobject.GObject, X11ServerBase):
                 except:
                     pass        #not in focus history!
             order[(prio, wid)] = window
-        log("do_repaint_root_overlay() has_focus=%s, has_grab=%s, windows in order=%s", self._has_focus, self._has_grab, order)
+        log("do_repaint_root_overlay() has_focus=%s, has_grab=%s, windows in order=%s",
+            self._has_focus, self._has_grab, order)
         for k in sorted(order):
             window = order[k]
             x, y, w, h = window.get_property("geometry")[:4]
@@ -1236,28 +1304,6 @@ class XpraServer(gobject.GObject, X11ServerBase):
         return False
 
 
-    def _handle_menu_rpc(self, ss, rpcid, action_type, wid, action, state, pdata, *extra):
-        if self.readonly:
-            return
-        menulog("_handle_menu_rpc%s", (ss, rpcid, action_type, wid, action, state, pdata, extra))
-        def native(args):
-            return [self.dbus_helper.dbus_to_native(x) for x in (args or [])]
-        def ok_back(*args):
-            menulog("menu rpc: ok_back%s", args)
-            ss.rpc_reply("menu", rpcid, True, native(args))
-        def err_back(*args):
-            menulog("menu rpc: err_back%s", args)
-            ss.rpc_reply("menu", rpcid, False, native(args))
-        try:
-            window = self._lookup_window(wid)
-            assert window, "window %i not found" % wid
-            window.activate_menu(action_type, action, state, pdata)
-            ok_back("done")
-        except Exception as e:
-            menulog.error("Error: cannot handle menu action %s:", action)
-            menulog.error(" %s", e)
-            err_back(str(e))
-
     def do_make_screenshot_packet(self):
         log("grabbing screenshot")
         regions = []
@@ -1275,7 +1321,7 @@ class XpraServer(gobject.GObject, X11ServerBase):
             try:
                 with xsync:
                     img = window.get_image(0, 0, w, h)
-            except:
+            except XError:
                 log.warn("screenshot: window %s could not be captured", wid)
                 continue
             if img is None:
